import xappt


@xappt.register_plugin
class StringUi(xappt.BaseTool):
    plain_string = xappt.ParamString()

    choice_string = xappt.ParamString(choices=("apple", "banana", "cantaloupe"))

    folder_select = xappt.ParamString(
        options={'ui': 'folder-select'},
        validators=[xappt.ValidateFolderExists])

    file_open = xappt.ParamString(
        options={'ui': 'file-open', "accept": ('Text Files *.txt', 'Images *.png;*.jpg;')},
        validators=[xappt.ValidateFileExists])

    file_save = xappt.ParamString(
        options={'ui': 'file-save', "accept": ('Text Files *.txt', 'All Files *.*')})

    multi_line = xappt.ParamString(options={'ui': 'multi-line'})

    label = xappt.ParamString(
        options={'ui': 'label'},
        value='Labels support <a href="https://www.w3.org/html/">HTML</a> with clickable links')

    markdown = xappt.ParamString(
        options={'ui': 'markdown'},
        value='Strings _also_ support [Markdown](https://daringfireball.net/projects/markdown/).\n\n'
              'Captions are **automatically** hidden when using the `label` or `markdown` ui.')

    password = xappt.ParamString(options={'ui': 'password'})

    csv_string = xappt.ParamString(
        options={
            'ui': 'csv',
            'header_row': True,
            'editable': True,
            'csv_import': True,
            'csv_export': True,
            'sorting_enabled': True
        },
        value=(
            'header1,header2,header3\n'
            'value1,value6,123\n'
            'value2,value5,12\n'
            'value3,value4,1234\n'
            'value4,value3,234\n'
            'value5,value2,3456\n'
            'value6,value1,12345\n'
        ))

    @classmethod
    def name(cls):
        return "string-ui"

    @classmethod
    def help(cls) -> str:
        return (
            "This is an example showing the many ui options for string widgets:\n\n"
            "* folder-select\n"
            "* file-open\n"
            "* file-save\n"
            "* multi-line\n"
            "* label\n"
            "* markdown\n"
            "* password\n"
            "* csv\n"
        )

    @classmethod
    def collection(cls) -> str:
        return "Examples"

    def execute(self, **kwargs) -> int:
        return 0



console_word_wrap = False
console_auto_scroll = True

# https://doc.qt.io/qt-5/stylesheet-reference.html#paletterole
console_color_stdout = "palette(window-text)"
console_color_stderr = "#f55"


def load_settings():
    import json
    from xappt_qt.constants import APP_CONFIG_PATH
    config_path = APP_CONFIG_PATH.joinpath("settings.cfg")
    try:
        with open(config_path, "r") as fp:
            settings_raw = json.load(fp)
    except (FileNotFoundError, json.JSONDecodeError):
        pass
    else:
        global console_word_wrap
        console_word_wrap = settings_raw.get('console_word_wrap', False)
        global console_auto_scroll
        console_auto_scroll = settings_raw.get('console_auto_scroll', True)


load_settings()
del load_settings



import argparse
import os
import sys

import xappt
import xappt_qt
import xappt_qt.browser
import xappt_qt.launcher

from xappt_qt.constants import APP_INTERFACE_NAME


if getattr(xappt_qt, "__compiled__", None) is not None:
    xappt_qt.executable = os.path.abspath(sys.argv[0])
elif getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS"):
    xappt_qt.executable = os.path.abspath(sys.executable)


def main() -> int:
    os.environ[xappt.INTERFACE_ENV] = APP_INTERFACE_NAME

    parser = argparse.ArgumentParser()
    parser.add_argument('toolname', type=str, help='Specify the name of the tool to load', nargs='?')
    parser.add_argument('-v', '--version', action='store_true',
                        help='Display the version number and build')

    options, unknowns = parser.parse_known_args()

    if options.version:
        print(f"xappt_qt {xappt_qt.version_str}")

    if options.toolname is None:
        return xappt_qt.browser.main(sys.argv)
    else:
        return xappt_qt.launcher.launch(options.toolname, unknown_args=unknowns)


if __name__ == '__main__':
    sys.exit(main())



import time

import xappt


@xappt.register_plugin
class HeadlessExecution(xappt.BaseTool):
    headless = True

    @classmethod
    def name(cls) -> str:
        return "headless"

    @classmethod
    def help(cls) -> str:
        return ("A headless tool will run with no interface, except for a progress bar and "
                "message dialogs.\n\nTo make a tool headless, simply add a class variable "
                "named `headless`, and set its value to **True**.")

    @classmethod
    def collection(cls) -> str:
        return "Examples"

    def execute(self, **kwargs) -> int:
        self.interface.progress_start()
        for i in range(10):
            self.interface.progress_update(f"Working...", (i + 1) / 10)
            time.sleep(0.5)
        self.interface.progress_end()

        self.interface.message("Complete")

        return 0



import importlib.resources
import os
from typing import Generator

import pyperclip

from PyQt5 import QtGui, QtWidgets

from xappt_qt import config
from xappt_qt.gui.ui.console import Ui_Console


class ConsoleWidget(QtWidgets.QWidget, Ui_Console):
    STREAM_STDOUT = 0
    STREAM_STDERR = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setupUi(self)
        self.load_icons()
        self.set_tooltips()

        self.output_buffer_raw = []

        self.connect_signals()

        self.txtConsole.setTabChangesFocus(True)

        self.word_wrap: bool = config.console_word_wrap
        self.btnWordWrap.setChecked(self.word_wrap)
        self.on_wrap_toggled(self.word_wrap)

        self.auto_scroll: bool = config.console_auto_scroll
        self.btnScrollDown.setChecked(self.auto_scroll)
        self.on_scroll_toggled(self.auto_scroll)

    def load_icons(self):
        with importlib.resources.path("xappt_qt.resources.icons", "copy.svg") as copy_path:
            copy_icon = QtGui.QIcon()
            copy_icon.addPixmap(QtGui.QPixmap(str(copy_path)), QtGui.QIcon.Normal, QtGui.QIcon.On)
            self.btnCopy.setIcon(copy_icon)

        with importlib.resources.path("xappt_qt.resources.icons", "word-wrap.svg") as wrap_path:
            wrap_icon = QtGui.QIcon()
            wrap_icon.addPixmap(QtGui.QPixmap(str(wrap_path)), QtGui.QIcon.Normal, QtGui.QIcon.On)
            self.btnWordWrap.setIcon(wrap_icon)

        with importlib.resources.path("xappt_qt.resources.icons", "scroll-down.svg") as scroll_path:
            scroll_icon = QtGui.QIcon()
            scroll_icon.addPixmap(QtGui.QPixmap(str(scroll_path)), QtGui.QIcon.Normal, QtGui.QIcon.On)
            self.btnScrollDown.setIcon(scroll_icon)

        with importlib.resources.path("xappt_qt.resources.icons", "trash.svg") as trash_path:
            trash_icon = QtGui.QIcon()
            trash_icon.addPixmap(QtGui.QPixmap(str(trash_path)), QtGui.QIcon.Normal, QtGui.QIcon.On)
            self.btnTrash.setIcon(trash_icon)

    def set_tooltips(self):
        self.btnCopy.setToolTip("Copy to Clipboard")
        self.btnWordWrap.setToolTip("Word Wrap")
        self.btnScrollDown.setToolTip("Auto Scroll")
        self.btnTrash.setToolTip("Clear Output")

    def connect_signals(self):
        self.btnCopy.clicked.connect(self.on_copy)
        self.btnWordWrap.toggled.connect(self.on_wrap_toggled)
        self.btnScrollDown.toggled.connect(self.on_scroll_toggled)
        self.btnTrash.clicked.connect(self.on_clear)

    def on_copy(self):
        pyperclip.copy(os.linesep.join(self.output_buffer_raw))

    def on_wrap_toggled(self, state: bool):
        self.word_wrap = state
        if self.word_wrap:
            self.txtConsole.setLineWrapMode(QtWidgets.QTextEdit.WidgetWidth)
        else:
            self.txtConsole.setLineWrapMode(QtWidgets.QTextEdit.NoWrap)

    def on_scroll_toggled(self, state: bool):
        self.auto_scroll = state

    def on_clear(self):
        self.output_buffer_raw.clear()
        self.txtConsole.clear()

    def write_stdout(self, s: str):
        self._write_output(s, self.STREAM_STDOUT)

    def write_stderr(self, s: str):
        self._write_output(s, self.STREAM_STDERR)

    @staticmethod
    def convert_leading_whitespace(s: str, tabwidth: int = 4) -> str:
        leading_spaces = 0
        while True:
            if not len(s):
                break
            if s[0] == " ":
                leading_spaces += 1
            elif s[0] == "\t":
                leading_spaces += tabwidth
            else:
                break
            s = s[1:]
        return f"{'&nbsp;' * leading_spaces}{s}"

    def _write_output(self, s: str, stream: int):
        s = self.convert_leading_whitespace(s)
        color = config.console_color_stdout
        if stream == self.STREAM_STDERR:
            color = config.console_color_stderr
        html = f'<span style="color: {color}">{s}</span>'

        self.output_buffer_raw.append(s)
        self.txtConsole.append(html)

        if self.auto_scroll:
            self.txtConsole.moveCursor(QtGui.QTextCursor.End)
            self.txtConsole.moveCursor(QtGui.QTextCursor.StartOfLine)

    def ordered_widgets(self) -> Generator[QtWidgets.QWidget, None, None]:
        yield self.txtConsole
        yield self.btnCopy
        yield self.btnWordWrap
        yield self.btnScrollDown
        yield self.btnTrash



import base64
import os
import platform
import sys


from PyQt5 import QtWidgets, QtGui, QtCore

import xappt

from xappt_qt.gui.application import get_application
from xappt_qt.gui.ui.browser import Ui_Browser
from xappt_qt.gui.utilities.tray_icon import TrayIcon
from xappt_qt.constants import *
from xappt_qt.gui.tab_pages import ToolsTabPage, OptionsTabPage, AboutTabPage

from xappt_qt.utilities import singleton

DISABLE_TRAY_ICON = platform.system() == "Darwin"


class XapptBrowser(xappt.ConfigMixin, QtWidgets.QMainWindow, Ui_Browser):
    def __init__(self):
        super().__init__()

        self.app = get_application()

        self.setupUi(self)
        self.setWindowTitle(APP_TITLE)

        self.setWindowIcon(self.app.app_icon)
        self.tray_icon = TrayIcon(self, self.app.app_icon)

        self.tools = ToolsTabPage(on_info=self.tray_icon.info, on_warn=self.tray_icon.warn,
                                  on_error=self.tray_icon.critical)
        self.options = OptionsTabPage()
        self.about = AboutTabPage()

        self.tabWidget.addTab(self.tools, self.tools.windowTitle())
        self.tabWidget.addTab(self.options, self.options.windowTitle())
        self.tabWidget.addTab(self.about, self.about.windowTitle())
        self.tabWidget.setCurrentIndex(0)

        self.options.options_changed.connect(self.on_options_changed)
        if DISABLE_TRAY_ICON:
            self.options.disable_tray_icon()

        self.config_path = APP_CONFIG_PATH.joinpath("browser.cfg")
        self.init_config()
        self.load_config()

        self.init_tray_icon()

    def init_config(self):
        self.add_config_item('window-geo',
                             saver=self.save_window_geo,
                             loader=self.load_window_geo,
                             default=None)
        self.add_config_item('settings',
                             saver=lambda: self.options.settings(),
                             loader=lambda x: self.options.apply_settings(x),
                             default={})

    def load_window_geo(self, geo: str):
        try:
            self.restoreGeometry(QtCore.QByteArray(base64.b64decode(geo)))
        except TypeError:
            pass

    def save_window_geo(self) -> str:
        geo = bytes(self.saveGeometry())
        return base64.b64encode(geo).decode('utf8')

    def init_tray_icon(self):
        if DISABLE_TRAY_ICON:
            return

        for category, plugin_list in self.tools.loaded_plugins.items():
            for plugin in plugin_list:
                self.tray_icon.add_menu_item(plugin.name(), on_activate=self.run_tool, data=plugin, group=category)

        self.tray_icon.add_menu_item(None)
        self.tray_icon.add_menu_item("Show", on_activate=self.show, is_visible=self.isHidden)
        self.tray_icon.add_menu_item("Hide", on_activate=self.hide, is_visible=self.isVisible)
        self.tray_icon.add_menu_item(None)
        self.tray_icon.add_menu_item("Quit", on_activate=self.on_quit)
        self.tray_icon.on_trigger = self.on_activate
        self.tray_icon.show()

    def changeEvent(self, event: QtCore.QEvent):
        if not DISABLE_TRAY_ICON:
            if event.type() == QtCore.QEvent.WindowStateChange:
                if self.windowState() == QtCore.Qt.WindowMinimized:
                    if self.options.settings().get('minimize_to_tray') and self.tray_icon.tray_available:
                        self.hide()
        super().changeEvent(event)

    def closeEvent(self, event: QtGui.QCloseEvent):
        if self.options.settings().get('minimize_to_tray') and self.tray_icon.tray_available:
            event.ignore()
            self.hide()
        else:
            self.on_quit()

    def on_activate(self):
        self.show()
        self.raise_()
        self.activateWindow()

    def on_quit(self):
        self.tray_icon.destroy()
        self.save_config()
        self.app.quit()

    def run_tool(self, **kwargs):
        plugin = kwargs.get('data')
        if plugin is None:
            return
        assert issubclass(plugin, xappt.BaseTool)
        self.tools.launch_tool(plugin)

    def on_options_changed(self):
        settings = self.options.settings()
        self.tools.settings_changed(settings)
        self.options.settings_changed(settings)
        self.about.settings_changed(settings)


def main(args) -> int:
    try:
        _ = singleton.SingleInstance(flavor_id='browser')
    except singleton.SingleInstanceException:
        return 1

    app = get_application()

    browser = XapptBrowser()
    if browser.options.settings().get('start_minimized'):
        if not browser.options.settings().get('minimize_to_tray'):
            browser.setWindowState(QtCore.Qt.WindowMinimized)
            browser.show()
    else:
        browser.show()

    return app.exec_()


def entry_point() -> int:
    os.environ[xappt.INTERFACE_ENV] = APP_INTERFACE_NAME
    return main(sys.argv)


if __name__ == '__main__':
    sys.exit(entry_point())



import os

from . import dark_palette
from xappt_qt import constants


def apply_style(app):
    xappt_qt_stylesheet = os.getenv(constants.APP_STYLESHEET_ENV)
    if xappt_qt_stylesheet is not None and os.path.isfile(xappt_qt_stylesheet):
        with open(xappt_qt_stylesheet, 'r') as fp:
            custom_stylesheet = fp.read()
        app.setStyleSheet(custom_stylesheet)
    else:
        dark_palette.apply_palette(app)



import markdown


def to_markdown(text: str):
    if len(text) == 0:
        return text

    md = markdown.markdown(text)
    style = "".join((
        "code {background-color: #000; color: #ccc;}",
        "ul {margin-left: -20px;}",
        "a {text-decoration: none;}",
    ))
    wrap = f"<html><head><style>{style}</style></head><body>{md}</body></html>"
    return wrap



# -*- coding: utf-8 -*-


#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_ToolInterface(object):
    def setupUi(self, ToolInterface):
        ToolInterface.setObjectName("ToolInterface")
        ToolInterface.resize(775, 632)
        self.gridLayout = QtWidgets.QGridLayout(ToolInterface)
        self.gridLayout.setObjectName("gridLayout")
        self.splitter = QtWidgets.QSplitter(ToolInterface)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.splitter.setObjectName("splitter")
        self.toolContainer = QtWidgets.QWidget(self.splitter)
        self.toolContainer.setObjectName("toolContainer")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.toolContainer)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.consoleContainer = QtWidgets.QWidget(self.splitter)
        self.consoleContainer.setObjectName("consoleContainer")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.consoleContainer)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gridLayout.addWidget(self.splitter, 0, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.btnHelp = QtWidgets.QToolButton(ToolInterface)
        self.btnHelp.setIconSize(QtCore.QSize(24, 24))
        self.btnHelp.setAutoRaise(True)
        self.btnHelp.setObjectName("btnHelp")
        self.horizontalLayout.addWidget(self.btnHelp)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.btnRun = QtWidgets.QPushButton(ToolInterface)
        self.btnRun.setObjectName("btnRun")
        self.horizontalLayout.addWidget(self.btnRun)
        self.btnAdvance = QtWidgets.QPushButton(ToolInterface)
        self.btnAdvance.setObjectName("btnAdvance")
        self.horizontalLayout.addWidget(self.btnAdvance)
        self.btnRunAndAdvance = QtWidgets.QPushButton(ToolInterface)
        self.btnRunAndAdvance.setObjectName("btnRunAndAdvance")
        self.horizontalLayout.addWidget(self.btnRunAndAdvance)
        self.gridLayout.addLayout(self.horizontalLayout, 2, 0, 1, 1)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.progressBar = QtWidgets.QProgressBar(ToolInterface)
        self.progressBar.setProperty("value", 0)
        self.progressBar.setAlignment(QtCore.Qt.AlignCenter)
        self.progressBar.setFormat("")
        self.progressBar.setObjectName("progressBar")
        self.horizontalLayout_2.addWidget(self.progressBar)
        self.btnAbort = QtWidgets.QPushButton(ToolInterface)
        self.btnAbort.setAutoDefault(False)
        self.btnAbort.setObjectName("btnAbort")
        self.horizontalLayout_2.addWidget(self.btnAbort)
        self.gridLayout.addLayout(self.horizontalLayout_2, 1, 0, 1, 1)

        self.retranslateUi(ToolInterface)
        QtCore.QMetaObject.connectSlotsByName(ToolInterface)
        ToolInterface.setTabOrder(self.btnAbort, self.btnHelp)
        ToolInterface.setTabOrder(self.btnHelp, self.btnRun)
        ToolInterface.setTabOrder(self.btnRun, self.btnAdvance)
        ToolInterface.setTabOrder(self.btnAdvance, self.btnRunAndAdvance)

    def retranslateUi(self, ToolInterface):
        _translate = QtCore.QCoreApplication.translate
        ToolInterface.setWindowTitle(_translate("ToolInterface", "Dialog"))
        self.btnHelp.setText(_translate("ToolInterface", "?"))
        self.btnRun.setText(_translate("ToolInterface", "Run"))
        self.btnAdvance.setText(_translate("ToolInterface", "Next"))
        self.btnRunAndAdvance.setText(_translate("ToolInterface", "Next"))
        self.btnAbort.setText(_translate("ToolInterface", "Abort"))



from typing import Any, Callable, Dict, Generator, Optional, Type

from PyQt5 import QtWidgets, QtCore

import xappt

from xappt_qt.gui.widgets.tool_page.converters import *
from xappt_qt.gui.widgets.error_label import ErrorLabel


class ToolPage(QtWidgets.QWidget):
    def __init__(self, tool: xappt.BaseTool, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent=parent)

        self.convert_dispatch: Dict[Type, Callable] = {
            int: ParameterWidgetInt,
            bool: ParameterWidgetBool,
            float: ParameterWidgetFloat,
            str: ParameterWidgetStr,
            list: ParameterWidgetList,
        }

        self.vertical_expand = False

        self.tool = tool
        self.build_ui()

    # noinspection PyAttributeOutsideInit
    def build_ui(self):
        self.grid = QtWidgets.QGridLayout()
        self.grid.setColumnStretch(0, 0)
        self.grid.setColumnStretch(1, 1)
        self.grid.setColumnStretch(0, 0)
        self.grid.setHorizontalSpacing(16)
        self.grid.setVerticalSpacing(8)

        self.setLayout(self.grid)
        self._load_tool_parameters()

    def _convert_parameter(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        widget_class = self.convert_dispatch[param.data_type]
        widget_instance: ParameterWidgetBase = widget_class(parameter=param, parent=self)
        widget_instance.onValueChanged.connect(self.on_widget_value_changed)
        return widget_instance

    def _load_tool_parameters(self):
        index = 0
        for param in self.tool.parameters():
            if param.hidden:
                continue
            widget_instance = self._convert_parameter(param)
            if widget_instance.sizePolicy().verticalPolicy() == QtWidgets.QSizePolicy.Expanding:
                self.vertical_expand = True

            if len(widget_instance.caption):
                label = QtWidgets.QLabel(widget_instance.caption)
                label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
                label.setToolTip(param.description)
                param.metadata['label'] = label
                self.grid.addWidget(label, index, 0)

            error = ErrorLabel()
            param.metadata['error'] = error
            self.grid.addWidget(error, index, 2)

            widget_instance.setToolTip(param.description)
            param.metadata['widget'] = widget_instance  # this lets us avoid lambdas
            self.grid.addWidget(widget_instance, index, 1)

            self.parameter_options_updated(param)

            param.on_choices_changed.add(self.update_tool_choices)
            param.on_value_changed.add(self.parameter_value_updated)
            param.on_options_changed.add(self.parameter_options_updated)
            param.on_visibility_changed.add(self.parameter_visibility_updated)

            index += 1

    def ordered_widgets(self) -> Generator[QtWidgets.QWidget, None, None]:
        for row in range(self.grid.rowCount()):
            layout_item = self.grid.itemAtPosition(row, 1)
            if layout_item is None:
                continue
            widget = layout_item.widget()
            if widget is not None:
                yield widget

    def update_tool_choices(self, param: xappt.Parameter):
        """ Given that multiple parameter types can be updated at runtime,
        it's easier just to remove and recreate the widget rather than
        reimplementing a lot of the same functionality to update existing
        widgets. """
        widget = param.metadata.get('widget')
        if widget is None:
            return

        # find and remove existing widget
        index = self.grid.indexOf(widget)
        row, column, *_ = self.grid.getItemPosition(index)
        self.grid.takeAt(index)
        widget.deleteLater()

        # create a new widget to replace it
        new_widget = self._convert_parameter(param)

        new_widget.setToolTip(param.description)
        param.metadata['widget'] = new_widget
        self.grid.addWidget(new_widget, row, column)
        self.parameter_options_updated(param)

    def on_widget_value_changed(self, param_name: str, param_value: Any):
        parameter: xappt.Parameter = getattr(self.tool, param_name)
        error_widget: ErrorLabel = parameter.metadata["error"]

        try:
            parameter.value = parameter.validate(param_value)
        except xappt.ParameterValidationError as err:
            error_widget.set_error(str(err))
        else:
            error_widget.reset()

    @staticmethod
    def parameter_value_updated(param: xappt.Parameter):
        """ Update widget with the parameter's new value """
        widget = param.metadata.get('widget')
        assert widget is not None
        setter = param.metadata.get('ui-setter')
        assert setter is not None

        # Block signals to avoid infinite recursion:
        #    Updating the widget's value here can trigger the widget's onValueChanged
        #    signal, which would then update the parameter, cause its on_value_changed
        #    callback to be invoked, and then then trigger this function leading which
        #    would start the process over again.
        widget.blockSignals(True)
        setter(param.value)
        widget.blockSignals(False)

    @staticmethod
    def parameter_options_updated(param: xappt.Parameter):
        widget: Optional[QtWidgets.QWidget] = param.metadata.get('widget')

        enabled = param.option("enabled", True)
        if widget is not None:
            widget.setEnabled(enabled)

    @staticmethod
    def parameter_visibility_updated(param: xappt.Parameter):
        label: Optional[QtWidgets.QLabel] = param.metadata.get('label')
        widget: Optional[QtWidgets.QWidget] = param.metadata.get('widget')
        error: Optional[ErrorLabel] = param.metadata.get('error')

        if label is not None:
            label.setVisible(not param.hidden)
        if widget is not None:
            widget.setVisible(not param.hidden)
        if error is not None:
            error.setVisible(not param.hidden)

    def disconnect(self):
        for param in self.tool.parameters():
            param.on_value_changed.clear()
            param.on_options_changed.clear()
            param.on_choices_changed.clear()



from xappt_qt.gui.widgets.tool_page.converters.base import ParameterWidgetBase
from xappt_qt.gui.widgets.tool_page.converters.convert_bool import ParameterWidgetBool
from xappt_qt.gui.widgets.tool_page.converters.convert_float import ParameterWidgetFloat
from xappt_qt.gui.widgets.tool_page.converters.convert_int import ParameterWidgetInt
from xappt_qt.gui.widgets.tool_page.converters.convert_list import ParameterWidgetList
from xappt_qt.gui.widgets.tool_page.converters.convert_str import ParameterWidgetStr



from PyQt5 import QtWidgets

from typing import Callable


class BaseTabPage(QtWidgets.QWidget):
    def __init__(self, **kwargs):
        super().__init__()
        self._on_info: Callable = kwargs.get("on_info")
        self._on_warn: Callable = kwargs.get("on_warn")
        self._on_error: Callable = kwargs.get("on_error")

    def information(self, title: str, message: str):
        if self._on_info:
            self._on_info(title, message)
        else:
            print(f"{title}: {message}")

    def warning(self, title: str, message: str):
        if self._on_warn:
            self._on_warn(title, message)
        else:
            print(f"{title}: {message}")

    def critical(self, title: str, message: str):
        if self._on_error:
            self._on_error(title, message)
        else:
            print(f"{title}: {message}")

    def settings_changed(self, settings: dict):
        pass



from xappt_qt.plugins.tools.examples.video_convert import ConvertX265
from xappt_qt.plugins.tools.examples.headless import HeadlessExecution
from xappt_qt.plugins.tools.examples.custom_icon import *
from xappt_qt.plugins.tools.examples.auto_advance import AutoAdvance
from xappt_qt.plugins.tools.examples.string_ui import StringUi



from typing import Callable, Optional

from PyQt5 import QtCore, QtWidgets

import xappt


class ParameterWidgetBase(QtWidgets.QWidget):
    onValueChanged = QtCore.pyqtSignal(str, object)  # param name, param value

    def __init__(self, *, parameter: xappt.Parameter, parent: Optional[QtWidgets.QWidget]):
        super().__init__(parent=parent)

        self.caption = self._get_caption(parameter)
        self._getter_fn: Optional[Callable] = None
        self._setter_fn: Optional[Callable] = None

        w = self.get_widget(parameter)

        self.setSizePolicy(w.sizePolicy())
        layout = QtWidgets.QHBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        layout.addWidget(w)

        parameter.metadata['ui-setter'] = self._setter_fn

    @property
    def value(self):
        return self._getter_fn()

    @value.setter
    def value(self, value):
        self._setter_fn(value)

    def get_widget(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        raise NotImplementedError

    @staticmethod
    def _get_caption(param: xappt.Parameter) -> str:
        caption_default = param.name.replace("_", " ").title()
        caption = param.options.get("caption", caption_default)
        return caption

    @staticmethod
    def _setup_combobox(param: xappt.Parameter, combo_widget: QtWidgets.QComboBox):
        if param.options.get("searchable"):
            combo_widget.setEditable(True)
            combo_widget.setInsertPolicy(QtWidgets.QComboBox.NoInsert)
            combo_widget.lineEdit().editingFinished.connect(
                lambda: combo_widget.setCurrentIndex(combo_widget.findText(combo_widget.currentText())))
            completer = combo_widget.completer()
            completer.setCompletionMode(QtWidgets.QCompleter.PopupCompletion)
            completer.popup().setAlternatingRowColors(True)
            completer.setCaseSensitivity(QtCore.Qt.CaseInsensitive)
            completer.setFilterMode(QtCore.Qt.MatchContains)



import csv
import pathlib

from typing import Optional, Sequence

from PyQt5 import QtWidgets, QtCore, QtGui

from xappt_qt.constants import APP_TITLE


class CsvTextCapture:
    def __init__(self):
        self.rows = []

    def write(self, row: str):
        self.rows.append(row)


class TableEdit(QtWidgets.QTableWidget):
    data_changed = QtCore.pyqtSignal()

    COMMAND_INS_BEFORE = "Insert Before"
    COMMAND_INS_AFTER = "Insert After"
    COMMAND_DELETE = "Delete"
    COMMAND_RENAME = "Rename"

    COMMAND_IMPORT = "Import CSV..."
    COMMAND_EXPORT = "Export CSV..."

    CSV_FILTER = "CSV Files *.csv (*.csv)"
    IO_FILTERS = {  # filter text, suffix list
        CSV_FILTER: ".csv",
        "All Files * (*)": None,
    }

    def __init__(self, **kwargs):
        parent: Optional[QtWidgets.QWidget] = kwargs.get("parent")
        super().__init__(parent=parent)

        # attributes
        self._editable: bool = kwargs.get("editable", False)
        self._header_row: bool = kwargs.get("header_row", False)
        self._csv_import: bool = kwargs.get("csv_import", False)
        self._csv_export: bool = kwargs.get("csv_export", True)
        self._sorting_enabled: bool = kwargs.get("sorting_enabled", True)

        # setup
        self._init_table_attributes()
        self._init_context_menu()
        self._init_drag_and_drop()

        # state
        self._first_load = True
        self._last_path: pathlib.Path = pathlib.Path.cwd()
        self._last_filter: str = self.CSV_FILTER

        # signals
        self.itemChanged.connect(self.on_data_changed)

    def _init_table_attributes(self):
        if not self._editable:
            self.setEditTriggers(self.NoEditTriggers)

        self.setAlternatingRowColors(True)
        self.setSortingEnabled(self._sorting_enabled)

    def _init_drag_and_drop(self):
        self._acceptable_drop_suffix_list = []
        if self._csv_import:
            for suffix in self.IO_FILTERS.values():
                if isinstance(suffix, str):
                    self._acceptable_drop_suffix_list.append(suffix)
                elif isinstance(suffix, Sequence):
                    for item in suffix:
                        self._acceptable_drop_suffix_list.append(item)
            self.setAcceptDrops(True)

    def on_data_changed(self, _: QtWidgets.QTableWidgetItem):
        self.data_changed.emit()

    def _init_context_menu(self):
        if self._editable:
            self.horizontalHeader().setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            self.horizontalHeader().customContextMenuRequested.connect(self._on_context_menu_header_h)

            self.verticalHeader().setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            self.verticalHeader().customContextMenuRequested.connect(self._on_context_menu_header_v)

        if self._csv_import or self._csv_export:
            self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            self.customContextMenuRequested.connect(self._on_context_menu_table)

    def _on_context_menu_header_h(self, pos: QtCore.QPoint):
        column = self.columnAt(pos.x())
        self.selectColumn(column)

        menu_header = QtWidgets.QMenu()
        for action_name in (self.COMMAND_INS_BEFORE, self.COMMAND_INS_AFTER, self.COMMAND_DELETE, self.COMMAND_RENAME):
            action = QtWidgets.QAction(action_name, self)
            menu_header.addAction(action)

        action = menu_header.exec_(self.horizontalHeader().mapToGlobal(pos))
        if action is None:
            return
        if action.text() == self.COMMAND_INS_BEFORE:
            self.insertColumn(column)
        elif action.text() == self.COMMAND_INS_AFTER:
            self.insertColumn(column + 1)
        elif action.text() == self.COMMAND_DELETE:
            self.removeColumn(column)
        elif action.text() == self.COMMAND_RENAME:
            item = self.horizontalHeaderItem(column)
            if item is not None:
                current_name = item.text()
            else:
                current_name = ""
            new_name, success = QtWidgets.QInputDialog.getText(self, "Rename", "Rename Column", text=current_name)
            if success and len(new_name.strip()):
                self.horizontalHeaderItem(column).setText(new_name)
        else:
            return
        self.data_changed.emit()

    def _on_context_menu_header_v(self, pos: QtCore.QPoint):
        row = self.rowAt(pos.y())
        self.selectRow(row)

        menu_header = QtWidgets.QMenu()
        for action_name in (self.COMMAND_INS_BEFORE, self.COMMAND_INS_AFTER, self.COMMAND_DELETE):
            action = QtWidgets.QAction(action_name, self)
            menu_header.addAction(action)

        action = menu_header.exec_(self.verticalHeader().mapToGlobal(pos))
        if action is None:
            return
        if action.text() == self.COMMAND_INS_BEFORE:
            self.insertRow(row)
        elif action.text() == self.COMMAND_INS_AFTER:
            self.insertRow(row + 1)
        elif action.text() == self.COMMAND_DELETE:
            self.removeRow(row)
        else:
            return
        self.data_changed.emit()

    def _on_context_menu_table(self, pos: QtCore.QPoint):
        menu_table = QtWidgets.QMenu()

        if self._csv_import:
            menu_table.addAction(QtWidgets.QAction(self.COMMAND_IMPORT, self))

        if self._csv_export:
            menu_table.addAction(QtWidgets.QAction(self.COMMAND_EXPORT, self))

        action = menu_table.exec_(self.viewport().mapToGlobal(pos))
        if action is None:
            return
        if action.text() == self.COMMAND_IMPORT:
            self.prompt_open_file()
        elif action.text() == self.COMMAND_EXPORT:
            self.export_file()

    # noinspection PyPep8Naming
    def setText(self, source: str, **kwargs):
        header_row = kwargs.get("header_row", self._header_row)

        self.blockSignals(True)
        reader = csv.reader(source.splitlines())

        self.clear()

        rows = []
        column_count = 0
        for row in reader:
            rows.append(row)
            column_count = max(len(row), column_count)

        if column_count > 0 and len(rows):
            headers = rows.pop(0) if header_row else None

            self.setColumnCount(column_count)
            self.setRowCount(len(rows))
            if headers is not None:
                self.setHorizontalHeaderLabels(headers)

            for r, row in enumerate(rows):
                for c, text in enumerate(row):
                    item = QtWidgets.QTableWidgetItem(text)
                    self.setItem(r, c, item)

        if self._first_load:
            self.resizeColumnsToContents()
            self._first_load = False

        self.blockSignals(False)

    def text(self, **kwargs) -> str:
        rows = []

        include_headers = kwargs.get("include_headers", self._header_row)

        if include_headers:
            header_row = []
            for column in range(self.columnCount()):
                item = self.horizontalHeaderItem(column)
                header_row.append("" if item is None else item.text())
            rows.append(header_row)

        for row in range(self.rowCount()):
            this_row = []
            for column in range(self.columnCount()):
                item = self.item(row, column)
                this_row.append("" if item is None else item.text())
            rows.append(this_row)

        csv_capture = CsvTextCapture()
        csv.writer(csv_capture).writerows(rows)

        return "".join(csv_capture.rows)

    def prompt_open_file(self):
        file_name, selected_filter = QtWidgets.QFileDialog.getOpenFileName(
            self, 'Import File', str(self._last_path), ";;".join(self.IO_FILTERS.keys()), self._last_filter)

        if len(file_name) == 0:
            return

        header_row = self.ask("Does this file include a header row?", default=True)

        path = pathlib.Path(file_name)

        self._last_filter = selected_filter
        self._last_path = path.parent

        self._read_from_file(path, header_row=header_row)

    def export_file(self):
        file_name, selected_filter = QtWidgets.QFileDialog.getSaveFileName(
            self, 'Export File', str(self._last_path), ";;".join(self.IO_FILTERS.keys()), self._last_filter)

        if len(file_name) == 0:
            return

        include_headers = self.ask("Include headers in export?", default=True)

        path = pathlib.Path(file_name)

        self._last_filter = selected_filter
        self._last_path = path.parent

        suffix = self.IO_FILTERS[selected_filter]
        if suffix is None:
            # no action
            pass
        elif isinstance(suffix, str):
            if path.suffix.lower() != suffix:
                path = path.with_suffix(suffix)
        elif isinstance(suffix, Sequence):
            if path.suffix.lower() not in suffix:
                path = path.with_suffix(suffix[0])

        self._write_to_file(path, include_headers=include_headers)

        self.message(f"File saved: {path}")

    def _read_from_file(self, path: pathlib.Path, *, header_row: bool):
        if not path.is_file():
            raise FileNotFoundError(f"File does not exist: {path}")

        with path.open("r") as fp:
            contents = fp.read()

        self._first_load = True
        self.setText(contents, header_row=header_row)
        self.data_changed.emit()

    def _write_to_file(self, path: pathlib.Path, *, include_headers: bool):
        with path.open("w", newline="\n") as fp:
            fp.write(self.text(include_headers=include_headers))

    def message(self, message: str):
        QtWidgets.QMessageBox.information(self, APP_TITLE, message)

    def ask(self, message: str, *, default: bool = False) -> bool:
        if default:
            default_button = QtWidgets.QMessageBox.Yes
        else:
            default_button = QtWidgets.QMessageBox.No
        buttons = QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No
        ask_result = QtWidgets.QMessageBox.question(self, APP_TITLE, message, buttons=buttons,
                                                    defaultButton=default_button)
        return ask_result == QtWidgets.QMessageBox.Yes

    def dragEnterEvent(self, event: QtGui.QDragEnterEvent):
        urls = event.mimeData().urls()
        if len(urls) == 0:
            return
        drag_file = pathlib.Path(urls[0].toLocalFile())
        if not drag_file.is_file():
            return
        if drag_file.suffix.lower() in self._acceptable_drop_suffix_list:
            event.accept()

    def dragMoveEvent(self, event: QtGui.QDragMoveEvent):
        pass  # seems required for dropEvent to be called

    def dropEvent(self, event: QtGui.QDropEvent):
        urls = event.mimeData().urls()
        if len(urls) == 0:
            return
        drag_file = pathlib.Path(urls[0].toLocalFile())
        if not drag_file.is_file():
            return
        header_row = self.ask("Does this file include a header row?", default=True)
        self._read_from_file(drag_file, header_row=header_row)



import xappt


@xappt.register_plugin
class CustomIconResource(xappt.BaseTool):
    custom_icon = "xappt_qt.plugins.tools.examples.custom_icon::icon1.svg"

    @classmethod
    def name(cls) -> str:
        return "custom-icon-resource"

    @classmethod
    def help(cls) -> str:
        return ("This tool will display a custom icon in the **xappt_qt browser** and the **Tool dialog**.\n "
                "A class variable named `custom_icon` is set with a string value representing a module path "
                "and a file name bundled with that module.\n This information will be used with "
                "`importlib.resources` to retrieve the image's file path.\n\n"
                "*Note that the tool dialog will only display the custom icon of the first loaded tool.*")

    @classmethod
    def collection(cls) -> str:
        return "Examples"

    def execute(self, **kwargs) -> int:
        self.interface.message("Complete")
        return 0



# -*- coding: utf-8 -*-


#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_tabTools(object):
    def setupUi(self, tabTools):
        tabTools.setObjectName("tabTools")
        tabTools.resize(589, 600)
        self.gridLayout_2 = QtWidgets.QGridLayout(tabTools)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setContentsMargins(6, 6, 6, 6)
        self.gridLayout.setSpacing(12)
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSpacing(4)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.txtSearch = QtWidgets.QLineEdit(tabTools)
        self.txtSearch.setMinimumSize(QtCore.QSize(0, 24))
        self.txtSearch.setObjectName("txtSearch")
        self.horizontalLayout.addWidget(self.txtSearch)
        self.btnClear = QtWidgets.QToolButton(tabTools)
        self.btnClear.setMinimumSize(QtCore.QSize(0, 24))
        self.btnClear.setAutoRaise(True)
        self.btnClear.setObjectName("btnClear")
        self.horizontalLayout.addWidget(self.btnClear)
        self.gridLayout.addLayout(self.horizontalLayout, 0, 0, 1, 1)
        self.labelHelp = QtWidgets.QLabel(tabTools)
        self.labelHelp.setText("")
        self.labelHelp.setWordWrap(True)
        self.labelHelp.setObjectName("labelHelp")
        self.gridLayout.addWidget(self.labelHelp, 2, 0, 1, 1)
        self.treeTools = QtWidgets.QTreeWidget(tabTools)
        self.treeTools.setAlternatingRowColors(True)
        self.treeTools.setIndentation(0)
        self.treeTools.setRootIsDecorated(False)
        self.treeTools.setAnimated(True)
        self.treeTools.setAllColumnsShowFocus(True)
        self.treeTools.setObjectName("treeTools")
        self.treeTools.headerItem().setText(0, "1")
        self.treeTools.header().setVisible(False)
        self.gridLayout.addWidget(self.treeTools, 1, 0, 1, 1)
        self.gridLayout_2.addLayout(self.gridLayout, 0, 0, 1, 1)

        self.retranslateUi(tabTools)
        self.btnClear.clicked.connect(self.txtSearch.clear)
        QtCore.QMetaObject.connectSlotsByName(tabTools)
        tabTools.setTabOrder(self.txtSearch, self.btnClear)
        tabTools.setTabOrder(self.btnClear, self.treeTools)

    def retranslateUi(self, tabTools):
        _translate = QtCore.QCoreApplication.translate
        tabTools.setWindowTitle(_translate("tabTools", "Tools"))
        self.txtSearch.setPlaceholderText(_translate("tabTools", "Search"))
        self.btnClear.setText(_translate("tabTools", "x"))
        self.treeTools.setSortingEnabled(False)



# -*- coding: utf-8 -*-


#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Console(object):
    def setupUi(self, Console):
        Console.setObjectName("Console")
        Console.resize(755, 461)
        self.horizontalLayout = QtWidgets.QHBoxLayout(Console)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.txtConsole = QtWidgets.QTextEdit(Console)
        self.txtConsole.setUndoRedoEnabled(False)
        self.txtConsole.setReadOnly(True)
        self.txtConsole.setTabStopWidth(40)
        self.txtConsole.setTextInteractionFlags(QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.txtConsole.setObjectName("txtConsole")
        self.horizontalLayout.addWidget(self.txtConsole)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setContentsMargins(6, 6, 6, 6)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setObjectName("verticalLayout")
        self.btnCopy = QtWidgets.QToolButton(Console)
        self.btnCopy.setEnabled(True)
        self.btnCopy.setIconSize(QtCore.QSize(24, 24))
        self.btnCopy.setAutoRaise(True)
        self.btnCopy.setObjectName("btnCopy")
        self.verticalLayout.addWidget(self.btnCopy)
        self.btnWordWrap = QtWidgets.QToolButton(Console)
        self.btnWordWrap.setIconSize(QtCore.QSize(24, 24))
        self.btnWordWrap.setCheckable(True)
        self.btnWordWrap.setAutoRaise(True)
        self.btnWordWrap.setObjectName("btnWordWrap")
        self.verticalLayout.addWidget(self.btnWordWrap)
        self.btnScrollDown = QtWidgets.QToolButton(Console)
        self.btnScrollDown.setIconSize(QtCore.QSize(24, 24))
        self.btnScrollDown.setCheckable(True)
        self.btnScrollDown.setChecked(True)
        self.btnScrollDown.setAutoRaise(True)
        self.btnScrollDown.setObjectName("btnScrollDown")
        self.verticalLayout.addWidget(self.btnScrollDown)
        self.btnTrash = QtWidgets.QToolButton(Console)
        self.btnTrash.setIconSize(QtCore.QSize(24, 24))
        self.btnTrash.setAutoRaise(True)
        self.btnTrash.setObjectName("btnTrash")
        self.verticalLayout.addWidget(self.btnTrash)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.horizontalLayout.addLayout(self.verticalLayout)

        self.retranslateUi(Console)
        QtCore.QMetaObject.connectSlotsByName(Console)

    def retranslateUi(self, Console):
        _translate = QtCore.QCoreApplication.translate
        Console.setWindowTitle(_translate("Console", "Console"))
        self.btnCopy.setText(_translate("Console", "..."))
        self.btnWordWrap.setText(_translate("Console", "..."))
        self.btnScrollDown.setText(_translate("Console", "..."))
        self.btnTrash.setText(_translate("Console", "..."))



import time

import xappt


@xappt.register_plugin
class AutoAdvance(xappt.BaseTool):
    message = xappt.ParamString(options={"ui": "label"})
    next_iteration_advance_mode = xappt.ParamInt(choices=("no auto advance", "auto advance"))

    def __init__(self, *, interface: xappt.BaseInterface, **kwargs):
        super(AutoAdvance, self).__init__(interface=interface, **kwargs)

        self.max_iterations = 5
        self.auto_advance = bool(self.interface.tool_data.get('next_iteration_advance_mode', 0))

        self.iteration = self.interface.tool_data.get("iteration", 1)
        if self.iteration == self.max_iterations:
            step = "last"
            self.next_iteration_advance_mode.hidden = True
        else:
            step = xappt.humanize_ordinal(self.iteration)

        self.message.value = f"This is the {step} of {self.max_iterations} iterations of this tool."

    @classmethod
    def name(cls) -> str:
        return 'auto-advance'

    @classmethod
    def help(cls) -> str:
        return ("When using a tool in the Qt interface, the default behavior is to leave the tool disabled "
                "after a successful execution. Clicking **Next** or **Close** will move to the next tool or "
                "close the interface.\n\nYou can set an attribute named `auto_advance`, and when it's set "
                "to `True` the next tool will be automatically loaded (or the interface wil be closed) after "
                "a successful execution.")

    def message_label_text(self) -> str:
        raise NotImplementedError

    @classmethod
    def collection(cls) -> str:
        return "Examples"

    def execute(self, **kwargs) -> int:
        self.interface.progress_start()
        for i in range(100):
            progress = (i + 1) / 100.0
            self.interface.progress_update(f"Iteration: {i + 1}/100", progress)
            time.sleep(0.01)
        self.interface.progress_end()

        last_iteration = self.iteration == self.max_iterations
        if last_iteration:
            if self.auto_advance:
                self.interface.message("Auto Advance is enabled for this iteration.\n\nOnce you click OK "
                                       "on this message the interface will automatically exit.")
            else:
                self.interface.message("Auto Advance is disabled for this iteration.\n\nOnce you click OK "
                                       "on this message, click the Close button to exit this interface.")
        else:
            self.interface.add_tool(AutoAdvance)
            if self.auto_advance:
                self.interface.message("Auto Advance is enabled for this iteration.\n\nOnce you click OK "
                                       "on this message the next iteration will be automatically loaded.")
            else:
                self.interface.message("Auto Advance is disabled for this iteration.\n\nOnce you click OK "
                                       "on this message click the Next button to continue to the next "
                                       "iteration.")

        self.interface.tool_data['iteration'] = self.iteration + 1
        self.interface.tool_data['next_iteration_advance_mode'] = self.next_iteration_advance_mode.value

        return 0



from PyQt5 import QtWidgets, QtGui, QtCore

from xappt_qt.gui.application import get_application


def center_widget(widget: QtWidgets.QWidget):
    app = get_application()
    cursor_pos = QtGui.QCursor.pos()
    screen = app.screenAt(cursor_pos)
    screen_rect = screen.availableGeometry()
    window_rect = QtCore.QRect(QtCore.QPoint(0, 0), widget.frameSize().boundedTo(screen_rect.size()))
    widget.resize(window_rect.size())
    widget.move(screen_rect.center() - window_rect.center())



import enum
import os
import xappt

from xappt_qt.utilities import safe_file_name


__all__ = [
    'APP_TITLE_ENV',
    'APP_ICON_ENV',
    'APP_STYLESHEET_ENV',
    'APP_INTERFACE_NAME',
    'APP_PACKAGE_NAME',
    'APP_TITLE',
    'APP_CONFIG_PATH',
    'ToolViewType',
]

# Environment variables
APP_TITLE_ENV = "XAPPT_QT_TITLE"
APP_ICON_ENV = "XAPPT_QT_ICON"
APP_STYLESHEET_ENV = "XAPPT_QT_STYLESHEET"

os.environ.setdefault(APP_TITLE_ENV, "Xappt QT")

APP_INTERFACE_NAME = "qt"
APP_TITLE = os.environ.get(APP_TITLE_ENV)
APP_PACKAGE_NAME = safe_file_name(APP_TITLE)
APP_CONFIG_PATH = xappt.user_data_path().joinpath(APP_PACKAGE_NAME)


class ToolViewType(enum.IntEnum):
    VIEW_SMALL_ICONS = 0
    VIEW_LARGE_ICONS = 1



import pathlib
import re
import shutil

from contextlib import contextmanager
from typing import Optional

import xappt

FFMPEG_PRESETS = ("ultrafast", "superfast", "veryfast", "faster", "fast", "medium", "slow", "slower", "veryslow")  # noqa
TC_LINE_RE = re.compile(r"^\s*(?P<label>[a-z0-9_-]+)\s*[:=]\s*(?P<tc>\d\d:\d\d:\d\d\.\d+).*?", re.I)
TIMECODE_RE = re.compile(r"(?P<hours>\d\d):(?P<minutes>\d\d):(?P<seconds>\d\d\.\d+)$")
EXTENSIONS = (".mp4", ".mkv")


def timecode_to_seconds(tc: str) -> float:
    match = TIMECODE_RE.match(tc)
    if match is None:
        return 0
    hours = float(match.group("hours")) * 60.0 * 60.0
    minutes = float(match.group("minutes")) * 60.0
    seconds = float(match.group("seconds"))
    return hours + minutes + seconds


def build_ffmpeg_command(source: pathlib.Path, crf: int, preset: str) -> tuple:
    destination = xappt.unique_path(source.with_stem(f"{source.stem}.x265"), delimiter="-", length=2,
                                    mode=xappt.UniqueMode.INTEGER)
    return ("-i", str(source), "-c:v", "libx265", "-crf", str(crf), "-preset", preset,
            "-c:a", "copy", str(destination), "-y", "-progress", "pipe:1")


# noinspection DuplicatedCode
@xappt.register_plugin
class ConvertX265(xappt.BaseTool):
    source = xappt.ParamString(options={"ui": "file-open", "accept": ('Video files *.mp4;*.mkv', )},
                               validators=[xappt.ValidateFileExists])
    crf = xappt.ParamInt(minimum=1, maximum=51, default=18)
    preset = xappt.ParamString(choices=FFMPEG_PRESETS, default="veryfast")

    @classmethod
    def name(cls) -> str:
        return "convert-x265"

    @classmethod
    def help(cls) -> str:
        return ("This is a fairly complex example tool that uses [FFMpeg](https://www.ffmpeg.org/) to convert "
                "a video using the *x265 (HEVC)* codec.\n\nThe following techniques are demonstrated:\n\n"
                "* using the interface's subprocess mechanism\n"
                "* setting up callbacks to monitor stdout and stderr\n"
                "* storing settings between sessions.")

    @classmethod
    def collection(cls) -> str:
        return "Examples"

    def __init__(self, *, interface: xappt.BaseInterface, **kwargs):
        """ We're overriding `__init__` to set up some instance variables and to call
        `load_config`. See the note in `init_config()`.

        Note that `init_config()` is called in `BasePlugin.__init__`, so don't forget to
        call `self.load_config()` _after_ `super().__init__()`
        """
        super().__init__(interface=interface, **kwargs)
        self.total_seconds: Optional[float] = None
        self.current_file: Optional[str] = None
        self.load_config()

    def init_config(self):
        """ Set up persistent data. We're looping through all parameters and setting up a
        loader and saver function for each, except for "source".

        During `execute` we're calling `self.save_config()` which stores all config items
        using the specified saver function, and likewise during `__init__` we're calling
        `self.load_config()` which uses the specified loader function.
        """
        super().init_config()
        for param in self.parameters():
            if param.name == 'source':
                continue
            self.add_config_item(param.name,
                                 saver=lambda p=param: getattr(p, "value"),
                                 loader=lambda x, p=param: setattr(p, "value", x),
                                 default=param.default)

    def handle_progress(self, message: str):
        tc_match = TC_LINE_RE.match(message)
        if tc_match is None:
            return

        if self.total_seconds is None:
            if tc_match.group("label") == "Duration":
                self.total_seconds = timecode_to_seconds(tc_match.group("tc"))
        else:
            if tc_match.group("label") == "out_time":
                progress = timecode_to_seconds(tc_match.group("tc")) / self.total_seconds
                self.interface.progress_update(f"encoding {self.current_file} {progress * 100.0:.2f}%...", progress)

    @contextmanager
    def progress_callbacks(self):
        """ This context manager is simply to ensure that the progress handler callbacks
        are removed when we are done with them. This way we don't have to worry about tools
        that have not yet been garbage collected modifying the progress bars
        unintentionally. """
        self.interface.progress_start()
        self.interface.on_write_stdout.add(self.handle_progress)
        self.interface.on_write_stderr.add(self.handle_progress)
        try:
            yield
        finally:
            self.interface.on_write_stderr.remove(self.handle_progress)
            self.interface.on_write_stdout.remove(self.handle_progress)
            self.interface.progress_end()
            self.flush_progress_deferred_ops()

    def flush_progress_deferred_ops(self):
        """ Make sure there's a call to "invoke" to trigger callback removals. Pausing the callback
        prevents the invoke from actually doing anything other than processing deferred operations.
        """
        self.interface.on_write_stderr.paused = True
        self.interface.on_write_stderr.invoke("")
        self.interface.on_write_stderr.paused = False

        self.interface.on_write_stdout.paused = True
        self.interface.on_write_stdout.invoke("")
        self.interface.on_write_stdout.paused = False

    def execute(self, **kwargs) -> int:
        ffmpeg_bin = shutil.which('ffmpeg')
        if ffmpeg_bin is None:
            self.interface.error("ffmpeg binary not found")
            return 1

        self.save_config()

        source_path = pathlib.Path(self.source.value)

        self.total_seconds = None
        self.current_file = source_path.stem

        command = (ffmpeg_bin,) + build_ffmpeg_command(source_path, self.crf.value, self.preset.value)
        self.interface.write_stdout(xappt.CommandRunner.command_sequence_to_string(command))

        with self.progress_callbacks():
            result = self.interface.run_subprocess(command)

        if result == 0:
            self.interface.message("Complete")
        else:
            self.interface.error(f"Error: process finished with error code {result}")

        return result



import importlib.resources
import sys

from typing import Optional

from PyQt5 import QtWidgets, QtCore

import xappt

from xappt_qt.constants import *
from xappt_qt.gui.application import get_application


class HeadlessInterface(xappt.BaseInterface):
    def __init__(self):
        super().__init__()
        self.app = get_application()

        self.progress_dialog = QtWidgets.QProgressDialog()

    def setup_progress_dialog(self):
        self.progress_dialog.setMinimumWidth(600)

        flags = self.progress_dialog.windowFlags()
        flags = flags & ~QtCore.Qt.WindowContextHelpButtonHint  # noqa
        self.progress_dialog.setWindowFlags(flags)

        self.progress_dialog.canceled.connect(self.abort)

    def invoke(self, plugin: xappt.BaseTool, **kwargs) -> int:
        self.progress_dialog.setWindowTitle(f"{plugin.name()} - {APP_TITLE}")
        return plugin.execute(**kwargs)

    def message(self, message: str):
        QtWidgets.QMessageBox.information(None, APP_TITLE, message)

    def warning(self, message: str):
        QtWidgets.QMessageBox.warning(None, APP_TITLE, message)

    def error(self, message: str, *, details: Optional[str] = None):
        QtWidgets.QMessageBox.critical(None, APP_TITLE, message)

    def ask(self, message: str) -> bool:
        buttons = QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No
        ask_result = QtWidgets.QMessageBox.question(None, APP_TITLE, message, buttons=buttons,
                                                    defaultButton=QtWidgets.QMessageBox.No)
        return ask_result == QtWidgets.QMessageBox.Yes

    def progress_start(self):
        self.setup_progress_dialog()
        self.progress_dialog.setValue(0)
        self.progress_dialog.setLabelText("")
        self.progress_dialog.show()
        self.app.processEvents()

    def progress_update(self, message: str, percent_complete: float):
        progress_value = int(100.0 * percent_complete)
        self.progress_dialog.setValue(progress_value)
        self.progress_dialog.setLabelText(message)
        self.app.processEvents()

    def progress_end(self):
        self.progress_dialog.setValue(0)
        self.progress_dialog.setLabelText("")
        self.progress_dialog.canceled.disconnect(self.abort)
        self.progress_dialog.close()
        self.app.processEvents()

    def run(self, **kwargs) -> int:
        return super().run(**kwargs)

    def abort(self):
        if self.command_runner.running:
            self.command_runner.abort()
        raise SystemExit("Aborted by user")



import os

from xappt_qt.__version__ import __version__, __build__

from xappt_qt.plugins.interfaces.qt import QtInterface

# suppress "qt.qpa.xcb: QXcbConnection: XCB error: 3 (BadWindow)"
os.environ['QT_LOGGING_RULES'] = '*.debug=false;qt.qpa.*=false'

version = tuple(map(int, __version__.split('.'))) + (__build__, )
version_str = f"{__version__}-{__build__}"

executable = None



from math import ceil

from PyQt5 import QtWidgets, QtCore


class SimpleItemDelegate(QtWidgets.QItemDelegate):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._hint_height = None

    def sizeHint(self, option: QtWidgets.QStyleOptionViewItem, index: QtCore.QModelIndex) -> QtCore.QSize:
        hint = super().sizeHint(option, index)
        if self._hint_height is None:
            self._hint_height = int(ceil(hint.height() * 1.5))
        hint.setHeight(self._hint_height)
        return hint



#! /usr/bin/env python3

# from tendo project
# https://github.com/pycontribs/tendo

import logging
from multiprocessing import Process
import os
import sys
import tempfile
import unittest
import warnings

from typing import Optional


if sys.platform != "win32":
    import fcntl


class SingleInstanceException(BaseException):
    pass


class SingleInstance(object):
    def __init__(self, flavor_id: Optional[str] = None, lockfile: Optional[str] = None):
        warnings.simplefilter("ignore", ResourceWarning)
        self.initialized = False
        if flavor_id is None and lockfile is None:
            logger.error("Either 'flavor_id' or 'lockfile' is required, quitting.")
            raise SingleInstanceException()
        if lockfile is not None:
            self.lockfile = lockfile
        else:
            basename = sys.argv[0]
            basename = os.path.abspath(basename)
            basename = os.path.splitext(basename)[0]
            basename = basename.replace("\\", "-").replace("/", "-").replace(":", "").strip("-")
            basename = f"{basename}-{flavor_id}.lock"
            self.lockfile = os.path.join(tempfile.gettempdir(), basename)

        logger.debug("SingleInstance lockfile: " + self.lockfile)
        if sys.platform == 'win32':
            try:
                # file already exists, we try to remove (in case previous
                # execution was interrupted)
                if os.path.exists(self.lockfile):
                    os.unlink(self.lockfile)
                self.fd = os.open(self.lockfile, os.O_CREAT | os.O_EXCL | os.O_RDWR)
            except OSError:
                _, e, _ = sys.exc_info()
                if e.errno == 13:
                    logger.error("Another instance is already running, quitting.")
                    raise SingleInstanceException()
                print(e.errno)
                raise
        else:  # non Windows
            self.fp = open(self.lockfile, 'w')
            self.fp.flush()
            try:
                fcntl.lockf(self.fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
            except IOError:
                logger.warning("Another instance is already running, quitting.")
                raise SingleInstanceException()
        self.initialized = True

    def __del__(self):
        if not self.initialized:
            return
        try:
            if sys.platform == 'win32':
                if hasattr(self, 'fd'):
                    os.close(self.fd)
                    os.unlink(self.lockfile)
            else:
                fcntl.lockf(self.fp, fcntl.LOCK_UN)
                # os.close(self.fp)
                if os.path.isfile(self.lockfile):
                    os.unlink(self.lockfile)
        except Exception as e:
            if logger:
                logger.warning(e)
            else:
                print("Unloggable error: %s" % e)
            sys.exit(-1)


def f(name):
    tmp = logger.level
    logger.setLevel(logging.CRITICAL)  # we do not want to see the warning
    try:
        me2 = SingleInstance(flavor_id=name)  # noqa
    except SingleInstanceException:
        sys.exit(-1)
    logger.setLevel(tmp)


class TestSingleton(unittest.TestCase):
    def test_1(self):
        me = SingleInstance(flavor_id="test-1")
        del me  # now the lock should be removed
        assert True

    def test_2(self):
        p = Process(target=f, args=("test-2",))
        p.start()
        p.join()
        # the called function should succeed
        assert p.exitcode == 0, "%s != 0" % p.exitcode

    def test_3(self):
        me = SingleInstance(flavor_id="test-3")  # noqa -- me should still kept
        p = Process(target=f, args=("test-3",))
        p.start()
        p.join()
        # the called function should fail because we already have another
        # instance running
        assert p.exitcode != 0, "%s != 0 (2nd execution)" % p.exitcode
        # note, we return -1 but this translates to 255 meanwhile we'll
        # consider that anything different from 0 is good
        p = Process(target=f, args=("test-3",))
        p.start()
        p.join()
        # the called function should fail because we already have another
        # instance running
        assert p.exitcode != 0, "%s != 0 (3rd execution)" % p.exitcode

    def test_4(self):
        lockfile = '/tmp/foo.lock'
        me = SingleInstance(lockfile=lockfile)
        assert me.lockfile == lockfile


logger = logging.getLogger("tendo.singleton")


if __name__ == "__main__":
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.DEBUG)
    unittest.main()



import importlib.resources
import pathlib

from typing import Type, Union

from xappt import BaseTool
from xappt_qt.utilities.text import to_markdown

ICONS_MODULE = "xappt_qt.resources.icons"
TOOL_ICON = "tool.svg"


def get_tool_icon(tool: Union[Type[BaseTool], BaseTool]) -> pathlib.Path:
    if hasattr(tool, "custom_icon") and isinstance(tool.custom_icon, str):
        if tool.custom_icon.count("::"):
            module_path, file_name = tool.custom_icon.split("::", maxsplit=1)
        else:
            module_path = ".".join(tool.__module__.split(".")[:-1])  # get module's parent
            file_name = tool.custom_icon
        try:
            with importlib.resources.path(module_path, file_name) as path:
                icon_path = path
        except FileNotFoundError:
            pass
        else:
            return icon_path

    with importlib.resources.path(ICONS_MODULE, TOOL_ICON) as path:
        return path


def is_headless(tool: Union[Type[BaseTool], BaseTool]) -> bool:
    return getattr(tool, "headless", False)  # default: False


def help_text(tool: Union[Type[BaseTool], BaseTool], **kwargs) -> str:
    process_markdown: bool = kwargs.get('process_markdown', True)
    include_name: bool = kwargs.get('include_name', False)

    if include_name:
        if process_markdown:
            tool_help = f"## {tool.name()}\n\n{tool.help()}"
        else:
            tool_help = f"{tool.name()}\n\n{tool.help()}"
    else:
        tool_help = tool.help()

    if process_markdown:
        return to_markdown(tool_help)
    else:
        return tool_help


def can_auto_advance(tool: Union[Type[BaseTool], BaseTool]) -> bool:
    return getattr(tool, "auto_advance", False)



from __future__ import annotations

from PyQt5 import QtWidgets

import xappt

from xappt_qt.gui.widgets.tool_page.converters.base import ParameterWidgetBase
from xappt_qt.gui.widgets.check_list import CheckList


class ParameterWidgetList(ParameterWidgetBase):
    def get_widget(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        w = CheckList(searchable=param.options.get('searchable', False))
        if param.choices is not None:
            w.add_items(param.choices)
        for v in (param.value, param.default):
            if v is not None:
                w.check_items(v)
                break
        param.value = list(w.checked_items())
        w.item_changed.connect(lambda: self.onValueChanged.emit(param.name, self.value))

        self._getter_fn = lambda widget=w: list(widget.checked_items())
        self._setter_fn = lambda value, widget=w: self.set_list_value(value, widget)

        return w

    @staticmethod
    def set_list_value(items: list[str], widget: CheckList):
        widget.uncheck_all()
        widget.check_items(items)



import importlib.resources
import os
import sys

from PyQt5 import QtWidgets, QtGui

from xappt_qt.constants import APP_ICON_ENV


class XapptQtApplication(QtWidgets.QApplication):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        custom_icon = os.environ.get(APP_ICON_ENV)
        if custom_icon is not None and os.path.isfile(custom_icon):
            self.app_icon = QtGui.QIcon(str(custom_icon))
        else:
            with importlib.resources.path("xappt_qt.resources.icons", "appicon.svg") as appicon:
                self.app_icon = QtGui.QIcon(str(appicon))

        self.setWindowIcon(self.app_icon)

        from xappt_qt.gui.utilities.style import apply_style
        apply_style(self)


def get_application() -> XapptQtApplication:
    instance = QtWidgets.QApplication.instance()
    return instance or XapptQtApplication(sys.argv[1:])



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# adapted from https://gist.github.com/lschmierer/443b8e21ad93e2a2d7eb

from PyQt5.QtGui import QPalette, QColor
from PyQt5.QtCore import Qt

PRIMARY_COLOR = QColor(32, 32, 32)
SECONDARY_COLOR = PRIMARY_COLOR.darker(150)
HIGHLIGHT_COLOR = QColor(128, 176, 73)
BRIGHT_COLOR = QColor(192, 192, 192)
DISABLED_COLOR = QColor(64, 64, 64)


# noinspection PyUnresolvedReferences
def apply_palette(app):
    dark_palette = QPalette()

    dark_palette.setColor(QPalette.Window, PRIMARY_COLOR)
    dark_palette.setColor(QPalette.WindowText, BRIGHT_COLOR)
    dark_palette.setColor(QPalette.Base, SECONDARY_COLOR)
    dark_palette.setColor(QPalette.AlternateBase, PRIMARY_COLOR)
    dark_palette.setColor(QPalette.ToolTipBase, BRIGHT_COLOR)
    dark_palette.setColor(QPalette.ToolTipText, BRIGHT_COLOR)
    dark_palette.setColor(QPalette.Text, BRIGHT_COLOR)
    dark_palette.setColor(QPalette.Button, PRIMARY_COLOR)
    dark_palette.setColor(QPalette.ButtonText, BRIGHT_COLOR)
    dark_palette.setColor(QPalette.BrightText, Qt.red)
    dark_palette.setColor(QPalette.Link, HIGHLIGHT_COLOR)
    dark_palette.setColor(QPalette.Highlight, HIGHLIGHT_COLOR)
    dark_palette.setColor(QPalette.HighlightedText, SECONDARY_COLOR)

    dark_palette.setColor(QPalette.Light, PRIMARY_COLOR)

    dark_palette.setColor(QPalette.Disabled, QPalette.WindowText, DISABLED_COLOR)
    dark_palette.setColor(QPalette.Disabled, QPalette.Text, DISABLED_COLOR)
    dark_palette.setColor(QPalette.Disabled, QPalette.ButtonText, DISABLED_COLOR)

    app.setPalette(dark_palette)
    app.setStyleSheet("""
        QToolTip { 
            color: palette(text); 
            background-color: palette(window); 
            border: 1px solid palette(highlight);
            border-radius: 2px;
            padding: 2px 8px;
            min-width: 300px;
        }

        QPlainTextEdit {
            border: 1px solid palette(base);
            border-radius: 3px;
            background-color: palette(base);
            padding: 1px;
        }

        QPlainTextEdit:focus {
            border-color: palette(highlight);
        }

        QHeaderView::section {
            padding: 8px 16px;
        }
    """)



# -*- coding: utf-8 -*-


#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Browser(object):
    def setupUi(self, Browser):
        Browser.setObjectName("Browser")
        Browser.resize(262, 454)
        self.centralwidget = QtWidgets.QWidget(Browser)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.tabWidget = QtWidgets.QTabWidget(self.centralwidget)
        self.tabWidget.setTabPosition(QtWidgets.QTabWidget.West)
        self.tabWidget.setObjectName("tabWidget")
        self.gridLayout_2.addWidget(self.tabWidget, 0, 0, 1, 1)
        Browser.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(Browser)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 262, 27))
        self.menubar.setObjectName("menubar")
        Browser.setMenuBar(self.menubar)

        self.retranslateUi(Browser)
        self.tabWidget.setCurrentIndex(-1)
        QtCore.QMetaObject.connectSlotsByName(Browser)

    def retranslateUi(self, Browser):
        _translate = QtCore.QCoreApplication.translate
        Browser.setWindowTitle(_translate("Browser", "Xappt Browser"))



import base64
import enum
import os

from typing import Optional

from PyQt5 import QtWidgets, QtGui, QtCore

import xappt

from xappt_qt.constants import *
from xappt_qt.gui.dialogs.tool_ui_dialog import ToolUI
from xappt_qt.plugins.interfaces.headless import HeadlessInterface
from xappt_qt.gui.application import get_application
from xappt_qt.utilities.tool_attributes import *

os.environ.setdefault('QT_STYLE_OVERRIDE', "Fusion")


class ToolState(enum.Enum):
    LOADED = 0
    RUNNING = 1
    ERROR = 2
    SUCCESS = 3


@xappt.register_plugin
class QtInterface(xappt.BaseInterface):
    def __init__(self):
        super().__init__()
        self.app = get_application()

        self.ui = ToolUI()

        self.__ui_close_event_orig = self.ui.closeEvent
        self.ui.closeEvent = self.close_event

        self.ui.btnRun.clicked.connect(self.on_run_tool)
        self.ui.btnAdvance.clicked.connect(self.on_next_tool)
        self.ui.btnRunAndAdvance.clicked.connect(self.on_run_and_advance)
        self.ui.btnHelp.clicked.connect(self.on_show_help)
        self.ui.btnAbort.clicked.connect(self.on_abort)

        self.on_write_stdout.add(self.ui.write_stdout)
        self.on_write_stderr.add(self.ui.write_stderr)

        self._tool_geo = {}

    def init_config(self):
        self.add_config_item(key="tool_geo",
                             saver=lambda: self._tool_geo,
                             loader=lambda geo: self._tool_geo.update(geo),
                             default=dict())

    def load_window_geo(self, tool_key: str):
        self.load_config()
        geo = self._tool_geo.get(tool_key)
        try:
            self.ui.restoreGeometry(QtCore.QByteArray(base64.b64decode(geo)))
        except TypeError:
            pass

    def save_window_geo(self, tool_key: str):
        geo = bytes(self.ui.saveGeometry())
        self._tool_geo[tool_key] = base64.b64encode(geo).decode('utf8')
        self.save_config()

    @classmethod
    def name(cls) -> str:
        return APP_INTERFACE_NAME

    def invoke(self, plugin: xappt.BaseTool, **kwargs) -> int:
        return plugin.execute(**kwargs)

    def message(self, message: str):
        QtWidgets.QMessageBox.information(self.ui, APP_TITLE, message)

    def warning(self, message: str):
        QtWidgets.QMessageBox.warning(self.ui, APP_TITLE, message)

    def error(self, message: str, *, details: Optional[str] = None):
        QtWidgets.QMessageBox.critical(self.ui, APP_TITLE, message)

    def ask(self, message: str) -> bool:
        buttons = QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No
        ask_result = QtWidgets.QMessageBox.question(self.ui, APP_TITLE, message, buttons=buttons,
                                                    defaultButton=QtWidgets.QMessageBox.No)
        return ask_result == QtWidgets.QMessageBox.Yes

    def progress_start(self):
        self.ui.progressBar.setRange(0, 100)
        self.ui.progressBar.setFormat("")
        self.app.processEvents()

    def progress_update(self, message: str, percent_complete: float):
        progress_value = int(100.0 * percent_complete)
        self.ui.progressBar.setValue(progress_value)
        self.ui.progressBar.setFormat(message)
        self.app.processEvents()

    def progress_end(self):
        self.ui.progressBar.setValue(0)
        self.ui.progressBar.setFormat("")
        self.app.processEvents()

    def load_tool_ui(self):
        tool_class = self.get_tool(self.current_tool_index)
        tool_instance = tool_class(interface=self, **self.tool_data)
        self.ui.load_tool(tool_instance)
        self.set_tool_state(ToolState.LOADED)

    def run(self, **kwargs) -> int:
        if not len(self._tool_chain):
            return 2

        self._current_tool_index = 0
        tool_class = self.get_tool(self.current_tool_index)

        icon_path = get_tool_icon(tool_class)
        self.ui.setWindowIcon(QtGui.QIcon(str(icon_path)))
        self.ui.setWindowTitle(f"{tool_class.name()} - {APP_TITLE}")

        if is_headless(tool_class):
            headless_interface = HeadlessInterface()
            headless_interface.add_tool(tool_class)
            return headless_interface.run(**kwargs)

        tool_geo_key = f"{tool_class.collection()}::{tool_class.name()}"

        self.load_window_geo(tool_geo_key)
        self.load_tool_ui()

        self.ui.exec()
        self.save_window_geo(tool_geo_key)
        return 0

    def close_event(self, event: QtGui.QCloseEvent):
        if self.command_runner.running:
            if self.ask("A process is currently running.\nDo you want to kill it?"):
                self.command_runner.abort()
                self.warning("The Process has been terminated.")
            event.ignore()
        else:
            self.__ui_close_event_orig(event)

    def on_run_tool(self):
        self.set_tool_state(ToolState.RUNNING)
        tool = self.ui.current_tool
        try:
            tool.validate()
        except xappt.ParameterValidationError as err:
            self.error(str(err))
            self.set_tool_state(ToolState.ERROR)
        else:
            result = self.invoke(tool, **self.tool_data)
            if result == 0:
                self.set_tool_state(ToolState.SUCCESS)
            else:
                self.set_tool_state(ToolState.ERROR)

    def on_next_tool(self):
        self._current_tool_index = self.current_tool_index + 1
        try:
            self.load_tool_ui()
        except IndexError:
            self.ui.close()

    def on_run_and_advance(self):
        self.set_tool_state(ToolState.RUNNING)
        tool = self.ui.current_tool
        try:
            tool.validate()
        except xappt.ParameterValidationError as err:
            self.error(str(err))
            self.set_tool_state(ToolState.ERROR)
        else:
            result = self.invoke(tool, **self.tool_data)
            if result != 0:
                self.set_tool_state(ToolState.ERROR)
                return
            self.on_next_tool()

    def set_tool_state(self, state: ToolState):
        tool = self.ui.current_tool
        auto_advance = can_auto_advance(tool)

        self.ui.btnRun.setVisible(not auto_advance)
        self.ui.btnAdvance.setVisible(not auto_advance)
        self.ui.btnRunAndAdvance.setVisible(auto_advance)
        self.ui.btnRun.setEnabled(False)
        self.ui.btnAdvance.setEnabled(False)
        self.ui.btnRunAndAdvance.setEnabled(False)

        self.ui.btnHelp.setEnabled(state != ToolState.RUNNING)
        self.ui.btnAbort.setEnabled(state == ToolState.RUNNING)
        self.ui.btnAbort.setVisible(state == ToolState.RUNNING)

        if auto_advance:
            self.set_tool_state_auto_advance(state)
        else:
            self.set_tool_state_no_advance(state)

    def set_tool_state_auto_advance(self, state: ToolState):
        last_tool = self.current_tool_index == self.tool_count - 1

        if state == ToolState.LOADED:
            self.ui.btnRunAndAdvance.setEnabled(True)
            self.ui.set_tool_enabled(True)
            if last_tool:
                self.ui.btnRunAndAdvance.setText("Run")
            else:
                self.ui.btnRunAndAdvance.setText("Next")
        elif state == ToolState.RUNNING:
            self.ui.set_tool_enabled(False)
            self.ui.btnRunAndAdvance.setEnabled(False)
        elif state == ToolState.ERROR:
            self.ui.set_tool_enabled(True)
            self.ui.btnRunAndAdvance.setEnabled(True)
        elif state == ToolState.SUCCESS:
            pass

    def set_tool_state_no_advance(self, state: ToolState):
        last_tool = self.current_tool_index == self.tool_count - 1

        if state == ToolState.LOADED:
            self.ui.btnRun.setEnabled(True)
            self.ui.set_tool_enabled(True)
            self.ui.set_tool_enabled(True)
        elif state == ToolState.RUNNING:
            self.ui.set_tool_enabled(False)
            self.ui.btnRun.setEnabled(False)
            self.ui.btnAdvance.setEnabled(False)
        elif state == ToolState.ERROR:
            self.ui.set_tool_enabled(True)
            self.ui.btnRun.setEnabled(True)
            self.ui.btnAdvance.setEnabled(False)
        elif state == ToolState.SUCCESS:
            self.ui.btnRun.setEnabled(False)
            self.ui.btnAdvance.setEnabled(True)
            if last_tool:
                self.ui.btnAdvance.setText("Close")

    def on_show_help(self):
        tool = self.ui.current_tool
        if tool is None:
            return
        message = help_text(tool)
        if len(message):
            self.message(message)

    def on_abort(self):
        self.abort()



# -*- coding: utf-8 -*-


#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_tabAbout(object):
    def setupUi(self, tabAbout):
        tabAbout.setObjectName("tabAbout")
        tabAbout.resize(346, 415)
        self.gridLayout = QtWidgets.QGridLayout(tabAbout)
        self.gridLayout.setContentsMargins(16, 16, 16, 16)
        self.gridLayout.setHorizontalSpacing(16)
        self.gridLayout.setVerticalSpacing(8)
        self.gridLayout.setObjectName("gridLayout")
        self.lblIcon = QtWidgets.QLabel(tabAbout)
        self.lblIcon.setText("")
        self.lblIcon.setObjectName("lblIcon")
        self.gridLayout.addWidget(self.lblIcon, 0, 0, 1, 1)
        self.lblTitle = QtWidgets.QLabel(tabAbout)
        font = QtGui.QFont()
        font.setPointSize(18)
        font.setBold(True)
        font.setWeight(75)
        self.lblTitle.setFont(font)
        self.lblTitle.setAlignment(QtCore.Qt.AlignCenter)
        self.lblTitle.setObjectName("lblTitle")
        self.gridLayout.addWidget(self.lblTitle, 0, 1, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(159, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 0, 2, 1, 1)
        self.txtAbout = QtWidgets.QTextBrowser(tabAbout)
        self.txtAbout.setReadOnly(True)
        self.txtAbout.setOpenExternalLinks(True)
        self.txtAbout.setObjectName("txtAbout")
        self.gridLayout.addWidget(self.txtAbout, 1, 0, 1, 3)

        self.retranslateUi(tabAbout)
        QtCore.QMetaObject.connectSlotsByName(tabAbout)

    def retranslateUi(self, tabAbout):
        _translate = QtCore.QCoreApplication.translate
        tabAbout.setWindowTitle(_translate("tabAbout", "About"))
        self.lblTitle.setText(_translate("tabAbout", "Xappt QT"))



from __future__ import annotations

import argparse
import os
import sys

import xappt
from xappt.config import log as logger

from xappt_qt.constants import *


def parse_unknowns_args(args: list[str], default_bool=True):
    kwarg_dict = {}
    previous_key: str = ""
    while len(args):
        this_item = args.pop(0)
        if this_item.startswith("--"):
            if previous_key:
                # we don't have a value for the last key, make it a bool
                kwarg_dict[previous_key] = default_bool
            previous_key = this_item[2:]
        else:
            if previous_key:
                kwarg_dict[previous_key] = this_item
                previous_key = ""
            else:
                logger.warning(f"Unhandled argument '{this_item}'")
    if previous_key:  # dangling key, make it a bool
        kwarg_dict[previous_key] = default_bool
    return kwarg_dict


def launch(tool_name: str, unknown_args: list[str]):
    tool_class = xappt.get_tool_plugin(tool_name)
    if tool_class is None:
        raise SystemExit(f"Tool {tool_name} not found.")

    interface = xappt.get_interface()
    interface.tool_data = parse_unknowns_args(unknown_args)
    interface.add_tool(tool_class)

    return interface.run()


def main(argv) -> int:
    os.environ[xappt.INTERFACE_ENV] = APP_INTERFACE_NAME

    parser = argparse.ArgumentParser()
    parser.add_argument('toolname', help='Specify the name of the tool to load')

    options, unknowns = parser.parse_known_args(args=argv)

    xappt.discover_plugins()

    return launch(options.toolname, unknowns)


def entry_point() -> int:
    return main(sys.argv[1:])


if __name__ == '__main__':
    sys.exit(entry_point())



import webbrowser

from PyQt5 import QtWidgets, QtCore

import xappt

from xappt_qt.gui.widgets.tool_page.converters.base import ParameterWidgetBase
from xappt_qt.gui.widgets.file_edit import FileEdit
from xappt_qt.gui.widgets.text_edit import TextEdit
from xappt_qt.gui.widgets.table_edit import TableEdit
from xappt_qt.utilities.text import to_markdown


class ParameterWidgetStr(ParameterWidgetBase):
    def get_widget(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        if param.choices is not None:
            return self._convert_str_choice(param)
        else:
            return self._convert_str_edit(param)

    def _convert_str_choice(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        w = QtWidgets.QComboBox()
        w.addItems(param.choices)
        for v in (param.value, param.default):
            if v is not None and v in param.choices:
                index = w.findText(v)
                w.setCurrentIndex(index)
                break
        else:
            param.value = w.currentText()

        self._setup_combobox(param, w)

        w.currentIndexChanged[str].connect(lambda x: self.onValueChanged.emit(param.name, x))

        self._getter_fn = w.currentText
        self._setter_fn = lambda s, widget=w: widget.setCurrentIndex(widget.findText(s))

        return w

    def _convert_str_edit(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        ui = param.options.get("ui")
        if ui == "folder-select":
            w = FileEdit(mode=FileEdit.MODE_CHOOSE_DIR)
            w.onSetFile.connect(lambda x: self.onValueChanged.emit(param.name, x))
        elif ui == "file-open":
            w = FileEdit(accept=param.options.get("accept"), mode=FileEdit.MODE_OPEN_FILE)
            w.onSetFile.connect(lambda x: self.onValueChanged.emit(param.name, x))
        elif ui == "file-save":
            w = FileEdit(accept=param.options.get("accept"), mode=FileEdit.MODE_SAVE_FILE)
            w.onSetFile.connect(lambda x: self.onValueChanged.emit(param.name, x))
        elif ui == "multi-line":
            w = TextEdit()
            w.editingFinished.connect(lambda widget=w: self.onValueChanged.emit(param.name, widget.text()))
        elif ui in ("label", "markdown"):
            w = QtWidgets.QLabel()
            w.setTextFormat(QtCore.Qt.RichText)
            w.setWordWrap(True)
            self.caption = ""
            w.linkActivated.connect(self.link_activated)
        elif ui == "csv":
            w = TableEdit(header_row=param.options.get("header_row", False),
                          editable=param.options.get("editable", False),
                          csv_import=param.options.get("csv_import", False),
                          csv_export=param.options.get("csv_export", True),
                          sorting_enabled=param.options.get("sorting_enabled", True))
            w.data_changed.connect(lambda widget=w: self.onValueChanged.emit(param.name, widget.text()))
        else:
            w = QtWidgets.QLineEdit()
            if ui == "password":
                w.setEchoMode(QtWidgets.QLineEdit.Password)
            w.editingFinished.connect(lambda widget=w: self.onValueChanged.emit(param.name, widget.text()))

        self._getter_fn = w.text
        if ui == "markdown":
            self._setter_fn = lambda t, widget=w: w.setText(to_markdown(t))
        else:
            self._setter_fn = w.setText

        for v in (param.value, param.default):
            if v is not None:
                self._setter_fn(v)
                break
        else:
            self._setter_fn("")

        return w

    @staticmethod
    def link_activated(url: str):
        webbrowser.open(url)



import importlib.resources
import math
import platform
import subprocess
import sys
import webbrowser

from PyQt5 import QtWidgets, QtGui, QtCore

from collections import defaultdict
from typing import DefaultDict, Generator, List, Tuple

import xappt

import xappt_qt
import xappt_qt.config
from xappt_qt.constants import APP_TITLE, ToolViewType
from xappt_qt.gui.ui.browser_tab_tools import Ui_tabTools
from xappt_qt.gui.delegates import SimpleItemDelegate, ToolItemDelegate
from xappt_qt.gui.tab_pages.base import BaseTabPage
from xappt_qt.utilities.tool_attributes import *


ICON_SIZES = {
    ToolViewType.VIEW_SMALL_ICONS: QtCore.QSize(24, 24),
    ToolViewType.VIEW_LARGE_ICONS: QtCore.QSize(48, 48),
}


class ToolsTabPage(BaseTabPage, Ui_tabTools):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.setupUi(self)

        self.set_tree_attributes()

        with importlib.resources.path("xappt_qt.resources.icons", "clear.svg") as path:
            self.btnClear.setIcon(QtGui.QIcon(str(path)))

        self.loaded_plugins: DefaultDict[str, List[Type[xappt.BaseTool]]] = defaultdict(list)
        self.populate_plugins()
        self.connect_signals()

    def set_tree_attributes(self):
        self.treeTools.setIconSize(QtCore.QSize(24, 24))
        self.treeTools.setItemDelegate(ToolItemDelegate())

    def populate_plugins(self):
        self.treeTools.clear()
        plugin_list: DefaultDict[str, List[Type[xappt.BaseTool]]] = defaultdict(list)

        for _, plugin_class in xappt.plugin_manager.registered_tools():
            collection = plugin_class.collection()
            plugin_list[collection].append(plugin_class)

        for collection in sorted(plugin_list.keys(), key=lambda x: x.lower()):
            collection_item = self._create_collection_item(collection)
            self.treeTools.insertTopLevelItem(self.treeTools.topLevelItemCount(), collection_item)
            for plugin in sorted(plugin_list[collection], key=lambda x: x.name().lower()):
                tool_item = self._create_tool_item(plugin)
                collection_item.addChild(tool_item)
                self.loaded_plugins[collection].append(plugin)
            collection_item.setExpanded(True)

    def connect_signals(self):
        self.treeTools.itemActivated.connect(self.item_activated)
        self.treeTools.itemSelectionChanged.connect(self.selection_changed)
        self.treeTools.clicked.connect(self.on_tree_item_clicked)

        self.txtSearch.textChanged.connect(self.on_filter_tools)
        # noinspection PyAttributeOutsideInit
        self.__txtSearch_keyPressEvent_orig = self.txtSearch.keyPressEvent
        self.txtSearch.keyPressEvent = self._filter_key_press

        self.labelHelp.linkActivated.connect(self.on_link_activated)

    def on_tree_item_clicked(self, index: QtCore.QModelIndex):
        if index.data(ToolItemDelegate.ROLE_ITEM_TYPE) != ToolItemDelegate.ITEM_TYPE_COLLECTION:
            return
        if self.treeTools.isExpanded(index):
            self.treeTools.collapse(index)
        else:
            self.treeTools.expand(index)

    @staticmethod
    def _create_collection_item(collection_name: str) -> QtWidgets.QTreeWidgetItem:
        item = QtWidgets.QTreeWidgetItem()
        item.setText(0, collection_name)
        item.setData(0, ToolItemDelegate.ROLE_TOOL_CLASS, None)
        item.setData(0, ToolItemDelegate.ROLE_ITEM_TYPE, ToolItemDelegate.ITEM_TYPE_COLLECTION)
        return item

    @staticmethod
    def _create_tool_item(tool_class: Type[xappt.BaseTool]) -> QtWidgets.QTreeWidgetItem:
        item = QtWidgets.QTreeWidgetItem()
        item.setText(0, tool_class.name())
        item.setToolTip(0, help_text(tool_class, process_markdown=True, include_name=True))
        item.setData(0, ToolItemDelegate.ROLE_TOOL_CLASS, tool_class)
        item.setData(0, ToolItemDelegate.ROLE_ITEM_TYPE, ToolItemDelegate.ITEM_TYPE_TOOL)
        item.setData(0, ToolItemDelegate.ROLE_ITEM_SEARCH_TEXT, f'{tool_class.name()}\n{tool_class.help()}')

        icon_path = get_tool_icon(tool_class)
        item.setIcon(0, QtGui.QIcon(str(icon_path)))

        return item

    def item_activated(self, item: QtWidgets.QTreeWidgetItem, column: int):
        item_type = item.data(column, ToolItemDelegate.ROLE_ITEM_TYPE)
        if item_type != ToolItemDelegate.ITEM_TYPE_TOOL:
            return
        tool_class: Type[xappt.BaseTool] = item.data(column, ToolItemDelegate.ROLE_TOOL_CLASS)
        self.launch_tool(tool_class)

    @staticmethod
    def launch_command(tool_name: str) -> Tuple:
        if xappt_qt.executable is not None:
            return xappt_qt.executable, tool_name
        return sys.executable, "-m", "xappt_qt.launcher", tool_name

    def launch_tool(self, tool_class: Type[xappt.BaseTool]):
        tool_name = tool_class.name()
        try:
            launch_command = self.launch_command(tool_name)
        except TypeError:
            self.critical(APP_TITLE, "Could not find executable")
        else:
            if platform.system() == "Windows":
                proc = subprocess.Popen(launch_command, creationflags=subprocess.CREATE_NEW_CONSOLE)
            else:
                proc = subprocess.Popen(launch_command)
            self.information(APP_TITLE, f"Launched {tool_name} (pid {proc.pid})")

    def selection_changed(self):
        selected_items = self.treeTools.selectedItems()
        if len(selected_items):
            self.labelHelp.setText(selected_items[0].toolTip(0))
        else:
            self.labelHelp.setText("")

    def _filter_key_press(self, event: QtGui.QKeyEvent):
        if event.key() == QtCore.Qt.Key_Escape:
            self.txtSearch.clear()
        self.__txtSearch_keyPressEvent_orig(event)

    def on_filter_tools(self, text: str):
        if len(text) == 0:
            # noinspection PyTypeChecker
            iterator = QtWidgets.QTreeWidgetItemIterator(self.treeTools, QtWidgets.QTreeWidgetItemIterator.All)
            while iterator.value():
                item = iterator.value()
                item.setHidden(False)
                iterator += 1
            return
        search_terms = [item.lower() for item in text.split(" ") if len(item)]
        self._filter_branch(search_terms, self.treeTools.invisibleRootItem())

    def _filter_branch(self, search_terms: List[str], parent: QtWidgets.QTreeWidgetItem) -> int:
        visible_children = 0
        for c in range(parent.childCount()):
            child = parent.child(c)
            item_type = child.data(0, ToolItemDelegate.ROLE_ITEM_TYPE)
            if item_type == ToolItemDelegate.ITEM_TYPE_TOOL:
                search_text = child.data(0, ToolItemDelegate.ROLE_ITEM_SEARCH_TEXT)
                visible_children += 1
                item_hidden = False
                for term in search_terms:
                    if term not in search_text:
                        item_hidden = True
                        break
                child.setHidden(item_hidden)
                if item_hidden:
                    visible_children -= 1
            elif item_type == ToolItemDelegate.ITEM_TYPE_COLLECTION:
                visible_children += self._filter_branch(search_terms, child)
            else:
                raise NotImplementedError
        parent.setHidden(visible_children == 0)
        return visible_children

    @staticmethod
    def on_link_activated(url: str):
        webbrowser.open(url)

    def settings_changed(self, settings: dict):
        view_type = settings.get('view_type', 0)
        self.treeTools.setIconSize(ICON_SIZES[view_type])

        # noinspection PyTypeChecker
        iterator = QtWidgets.QTreeWidgetItemIterator(self.treeTools, QtWidgets.QTreeWidgetItemIterator.All)
        while iterator.value():
            item = iterator.value()
            if item.data(0, ToolItemDelegate.ROLE_ITEM_TYPE) == ToolItemDelegate.ITEM_TYPE_TOOL:
                item.setData(0, ToolItemDelegate.ROLE_ICON_SIZE, ICON_SIZES[view_type])
            else:
                item.setData(0, ToolItemDelegate.ROLE_ICON_SIZE, ICON_SIZES[ToolViewType.VIEW_SMALL_ICONS])
            iterator += 1



from collections import namedtuple
from typing import Any, Callable, Dict, Optional, Union

from PyQt5 import QtWidgets, QtGui, QtCore
from PyQt5.QtWidgets import QSystemTrayIcon

MenuAction = namedtuple('TrayIconMenuItem', ('on_activate', 'is_visible', 'data', 'group'))


class TrayIcon(QtCore.QObject):
    def __init__(self, widget: QtWidgets.QWidget, icon: QtGui.QIcon, **kwargs):
        super().__init__()
        self.menu_actions: Dict[str, MenuAction] = {}

        self.on_trigger: Optional[Callable] = kwargs.get('on_trigger')
        self.on_double_click: Optional[Callable] = kwargs.get('on_double_click')
        self.on_middle_click: Optional[Callable] = kwargs.get('on_middle_click')
        self.on_message_click: Optional[Callable] = kwargs.get('on_message_click')

        self.context_menu = QtWidgets.QMenu()

        self.tray_available = QSystemTrayIcon.isSystemTrayAvailable()
        self.messages_available = QSystemTrayIcon.supportsMessages()

        self.tray_icon = QSystemTrayIcon(widget)
        self.tray_icon.setIcon(icon)
        self.tray_icon.setContextMenu(self.context_menu)

        self._connect_signals()

    def show(self):
        if not self.tray_available:
            return
        self._build_context_menu()
        self.tray_icon.show()

    def destroy(self):
        if not self.tray_available:
            return
        self.tray_icon.deleteLater()

    def add_menu_item(self, name: Optional[str], **kwargs):
        on_activate: Optional[Callable] = kwargs.get('on_activate')
        is_visible: Optional[Callable] = kwargs.get('is_visible')
        data: Optional[Any] = kwargs.get('data')
        group: Optional[str] = kwargs.get('group')

        if name is None:
            name = f"--{len(self.menu_actions):02d}"

        self.menu_actions[name] = MenuAction(on_activate, is_visible, data, group)

    def _message(self, title: str, message: str,
                 icon: Union[QSystemTrayIcon.MessageIcon, QtGui.QIcon] = QSystemTrayIcon.Information,
                 delay: int = 10000):
        if not self.tray_available:
            return
        if self.messages_available:
            self.tray_icon.showMessage(title, message, icon, delay)
        else:
            print(f"{title}: {message}")

    def info(self, title: str, message: str, delay: int = 10000):
        self._message(title, message, QSystemTrayIcon.Information, delay)

    def warn(self, title: str, message: str, delay: int = 10000):
        self._message(title, message, QSystemTrayIcon.Warning, delay)

    def critical(self, title: str, message: str, delay: int = 10000):
        self._message(title, message, QSystemTrayIcon.Critical, delay)

    # noinspection PyUnresolvedReferences
    def _connect_signals(self):
        if not self.tray_available:
            return
        self.context_menu.aboutToShow.connect(self._build_context_menu)
        self.context_menu.triggered.connect(self._on_context_menu_action)
        self.tray_icon.messageClicked.connect(self._on_message_clicked)
        self.tray_icon.activated.connect(self._on_activated)

    def _on_activated(self, reason: QSystemTrayIcon.ActivationReason):
        if reason == QSystemTrayIcon.DoubleClick:
            if self.on_double_click is not None:
                self.on_double_click()
        elif reason == QSystemTrayIcon.MiddleClick:
            if self.on_middle_click is not None:
                self.on_middle_click()
        elif reason == QSystemTrayIcon.Trigger:
            if self.on_trigger is not None:
                self.on_trigger()

    def _on_message_clicked(self):
        if self.on_message_click is not None:
            self.on_message_click()

    def _build_context_menu(self):
        self.context_menu.clear()
        submenus = {}
        for name, menu_item in self.menu_actions.items():
            if name.startswith("--"):
                self.context_menu.addSeparator()
            else:
                if menu_item.is_visible is None or menu_item.is_visible():
                    group = menu_item.group
                    if group is not None:
                        submenu = submenus.get(group)
                        if submenu is None:
                            submenu = self.context_menu.addMenu(group)
                            submenus[group] = submenu
                        submenu.addAction(name)
                    else:
                        self.context_menu.addAction(name)

    def _on_context_menu_action(self, action: QtWidgets.QAction):
        if action is None:
            return
        selected_command = action.text()
        menu_action = self.menu_actions[selected_command]
        on_activate = menu_action.on_activate
        if on_activate is not None:
            if menu_action.data is not None:
                on_activate(data=menu_action.data)
                return
            on_activate()



from PyQt5 import QtCore, QtGui, QtWidgets


class TextEdit(QtWidgets.QPlainTextEdit):
    editingFinished = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._changed = False
        self.setTabChangesFocus(True)
        self.textChanged.connect(self._handle_text_changed)
        self.setLineWrapMode(self.NoWrap)

    def focusOutEvent(self, event: QtGui.QFocusEvent):
        if self._changed:
            self.editingFinished.emit()
        super().focusOutEvent(event)

    def _handle_text_changed(self):
        self._changed = True

    def setTextChanged(self, state: bool = True):  # noqa
        self._changed = state

    def setText(self, text: str):  # noqa
        self.setPlainText(text)
        self._changed = False

    def text(self) -> str:
        return self.toPlainText()



import importlib.resources
from PyQt5 import QtWidgets, QtCore

from xappt_qt.constants import *


class ErrorLabel(QtWidgets.QLabel):
    def __new__(cls):
        instance = super().__new__(cls)
        with importlib.resources.path("xappt_qt.resources.icons", "error.svg") as error_path:
            error_link = f'<a href="#"><img src="{error_path}" /></a>'
        cls.error_link = error_link
        return instance

    def __init__(self):
        super().__init__()

        self._message: str = ""
        self.linkActivated.connect(self._on_link_activated)

        self.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)

        size_policy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Ignored)
        size_policy.setHorizontalStretch(0)
        size_policy.setVerticalStretch(0)
        self.setSizePolicy(size_policy)

        self.reset()

    def reset(self):
        self._message = ""
        self.setText("")

    def set_error(self, message: str):
        self._message = message
        self.setText(self.error_link)

    def show_error(self):
        if len(self._message):
            QtWidgets.QMessageBox.critical(self.parent(), APP_TITLE, self._message)

    def _on_link_activated(self, _: str):
        self.show_error()



from PyQt5 import QtCore, QtWidgets

import xappt

from xappt_qt.gui.widgets.tool_page.converters.base import ParameterWidgetBase


class ParameterWidgetBool(ParameterWidgetBase):
    def get_widget(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        if param.options.get("ui") == "button":
            w = QtWidgets.QPushButton(parent=self)
            w.setAutoDefault(False)
            w.setText(self._get_caption(param))
            w.clicked.connect(lambda: self.onValueChanged.emit(param.name, True))
            self.caption = ""

            self._getter_fn = lambda: True
            self._setter_fn = lambda x: True
        else:
            w = QtWidgets.QCheckBox(parent=self)
            w.setText(self._get_caption(param))
            self.caption = ""

            for v in (param.value, param.default):
                if v is not None:
                    w.setChecked(v)
                    break
            else:
                param.value = w.isChecked()
            w.stateChanged.connect(lambda x: self.onValueChanged.emit(param.name, x == QtCore.Qt.Checked))

            self._getter_fn = w.isChecked
            self._setter_fn = w.setChecked

        return w



import importlib.resources
from itertools import chain
from typing import Optional

from PyQt5 import QtWidgets, QtCore, QtGui

import xappt

from xappt_qt.gui.application import get_application
from xappt_qt.gui.ui.tool_interface import Ui_ToolInterface
from xappt_qt.gui.widgets.console import ConsoleWidget
from xappt_qt.gui.widgets.tool_page.widget import ToolPage


class ToolUI(QtWidgets.QDialog, Ui_ToolInterface):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.app = get_application()

        self.current_tool: Optional[xappt.BaseTool] = None

        self.setupUi(self)
        self.set_window_attributes()

        self.console = ConsoleWidget()
        self.setup_console()

    def set_tool_enabled(self, enabled: bool = True):
        self.toolContainer.setEnabled(enabled)

    def set_window_attributes(self):
        mw = QtWidgets.QMainWindow()
        flags = mw.windowFlags()
        mw.deleteLater()
        self.setWindowFlags(flags)
        self.setWindowIcon(self.app.app_icon)
        with importlib.resources.path("xappt_qt.resources.icons", "help.svg") as path:
            self.btnHelp.setIcon(QtGui.QIcon(str(path)))

    def setup_console(self):
        font_size = QtGui.QFontDatabase.systemFont(QtGui.QFontDatabase.GeneralFont).pointSizeF()
        mono_font = QtGui.QFontDatabase.systemFont(QtGui.QFontDatabase.FixedFont)
        mono_font.setPointSizeF(font_size)
        self.console.setFont(mono_font)
        self.consoleContainer.layout().addWidget(self.console)
        self.hide_console()

    def clear_loaded_tool(self):
        layout: QtWidgets.QVBoxLayout = self.toolContainer.layout()
        while layout.count():
            item = layout.takeAt(0)
            widget = item.widget()
            if widget is not None:
                widget.deleteLater()

    def load_tool(self, tool_instance: xappt.BaseTool):
        self.clear_loaded_tool()
        self.current_tool = tool_instance
        widget = ToolPage(self.current_tool)

        layout: QtWidgets.QVBoxLayout = self.toolContainer.layout()
        layout.addWidget(self.wrap_widget(widget))

        self.set_tab_order(widget)

    def set_tab_order(self, tool_widget: ToolPage):
        ui_widgets = [self.btnHelp, self.btnRun, self.btnAdvance, self.btnRunAndAdvance]
        first_widget: Optional[QtWidgets.QWidget] = None
        last_widget: Optional[QtWidgets.QWidget] = None
        for widget in chain(tool_widget.ordered_widgets(), self.console.ordered_widgets(), ui_widgets):
            if last_widget is None:
                first_widget = widget
            else:
                self.setTabOrder(last_widget, widget)
            last_widget = widget
        if first_widget is not None:
            first_widget.setFocus()

    @staticmethod
    def wrap_widget(widget: ToolPage) -> QtWidgets.QWidget:
        scroller = QtWidgets.QScrollArea()
        scroller.setWidgetResizable(True)

        container = QtWidgets.QWidget()
        layout = QtWidgets.QVBoxLayout()

        layout.addWidget(widget)

        if not widget.vertical_expand:
            layout.addItem(QtWidgets.QSpacerItem(1, 1, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding))

        container.setFocusPolicy(QtCore.Qt.NoFocus)
        container.setLayout(layout)
        scroller.setFocusPolicy(QtCore.Qt.NoFocus)
        scroller.setWidget(container)

        return scroller

    def show_console(self):
        if self.splitter.sizes()[1] > 0:
            return
        half_height = int(self.height() * 0.5)
        self.splitter.setSizes((half_height, half_height))

    def hide_console(self):
        self.splitter.setSizes((self.height(), 0))

    def write_stdout(self, text: str):
        self.show_console()
        for line in text.splitlines():
            self.console.write_stdout(line)
        self.app.processEvents()

    def write_stderr(self, text: str):
        self.show_console()
        for line in text.splitlines():
            self.console.write_stderr(line)
        self.app.processEvents()



import fnmatch
import os
import re

from PyQt5 import QtWidgets
from PyQt5 import QtCore
from PyQt5 import QtGui

FILTER_EXT_FILTER = re.compile(r"(?:\*?(?P<ext>\.[a-z0-9_.*-]+))")
FILTER_NAME_FILTER = re.compile(r"(?P<label>.*?)(?:\s?\*?(?P<ext>\.\S+))")


# noinspection PyPep8Naming
class FileEdit(QtWidgets.QWidget):
    # noinspection PyUnresolvedReferences
    onSetFile = QtCore.pyqtSignal(str)

    MODE_OPEN_FILE = 0
    MODE_SAVE_FILE = 1
    MODE_CHOOSE_DIR = 2

    def __init__(self, *, parent=None, accept=None, mode=MODE_OPEN_FILE):
        super(FileEdit, self).__init__(parent=parent)

        self._parent = parent
        self._mode = mode
        self._data = None
        self._accept = []
        self._filter = ''
        self._default_path = None

        self._hLayout = QtWidgets.QHBoxLayout()
        self._lineEdit = QtWidgets.QLineEdit()
        self._btnBrowse = QtWidgets.QToolButton()

        self._btnBrowse.setText('...')
        self._btnBrowse.setAcceptDrops(False)

        self._hLayout.setContentsMargins(0, 0, 0, 0)
        self._hLayout.addWidget(self._lineEdit)
        self._hLayout.addWidget(self._btnBrowse)

        self.setLayout(self._hLayout)

        self._lineEdit.setAcceptDrops(False)
        self._lineEdit.returnPressed.connect(self.onReturnPressed)
        self._lineEdit.editingFinished.connect(self.onReturnPressed)
        self._btnBrowse.clicked.connect(self.onBrowseButton)

        self.setAcceptDrops(True)

        self.setAccept(accept)

    def _set_file(self, filename):
        self._lineEdit.setText(filename)
        if os.path.isdir(filename):
            self._default_path = filename
        elif os.path.isfile(filename):
            self._default_path = os.path.dirname(filename)
        # noinspection PyUnresolvedReferences
        self.onSetFile.emit(filename)

    def setPlaceholderText(self, s):
        self._lineEdit.setPlaceholderText(s)

    def accept(self):
        return self._accept

    def browseButton(self):
        return self._btnBrowse.isVisible()

    def button(self):
        return self._btnBrowse

    def data(self):
        return self._data

    def defaultPath(self):
        return self._default_path

    def dragEnterEvent(self, event: QtGui.QDragEnterEvent):
        drag_file = event.mimeData().urls()[0].toLocalFile()

        if self._mode == self.MODE_CHOOSE_DIR:
            if os.path.isdir(drag_file):
                event.accept()
            return

        if not os.path.isfile(drag_file):
            return

        if len(self._accept):
            for ext in self._accept:
                if fnmatch.fnmatch(drag_file, f"*{ext}"):
                    event.accept()
                    return
        else:
            event.accept()

    def dropEvent(self, event: QtGui.QDropEvent):
        drag_file = event.mimeData().urls()[0].toLocalFile()
        self._set_file(drag_file)

    def filter(self):
        return self._filter

    def lineEdit(self):
        return self._lineEdit

    def _browse_file(self, starting_directory):
        if self._mode == self.MODE_OPEN_FILE:
            file_name, _ = QtWidgets.QFileDialog.getOpenFileName(
                self._parent, 'Select File', starting_directory, self._filter)
        elif self._mode == self.MODE_SAVE_FILE:
            file_name, _ = QtWidgets.QFileDialog.getSaveFileName(
                self._parent, 'Select File', starting_directory, self._filter,
                options=QtWidgets.QFileDialog.DontConfirmOverwrite)
        else:
            raise NotImplementedError
        if len(file_name) == 0:
            return
        self._set_file(file_name)

    def _browse_dir(self, starting_directory):
        file_name = QtWidgets.QFileDialog.getExistingDirectory(
            self._parent, 'Select Folder', starting_directory,
            QtWidgets.QFileDialog.ShowDirsOnly | QtWidgets.QFileDialog.DontResolveSymlinks)
        if file_name is not None and os.path.isdir(file_name):
            self._set_file(file_name)

    def onBrowseButton(self):
        current_path = self._lineEdit.text()
        if os.path.isdir(current_path):
            starting_directory = current_path
        elif os.path.isdir(os.path.dirname(current_path)):
            starting_directory = os.path.dirname(current_path)
        else:
            starting_directory = self._default_path

        if self._mode == self.MODE_CHOOSE_DIR:
            self._browse_dir(starting_directory)
        else:
            self._browse_file(starting_directory)

    def onReturnPressed(self):
        self._set_file(self._lineEdit.text())

    def setAccept(self, accept):
        self._accept = []
        filter_list = []
        for item in accept or []:
            extensions = FILTER_EXT_FILTER.findall(item)
            if len(extensions) == 0:
                continue
            name_match = FILTER_NAME_FILTER.match(item)
            if name_match is None:
                continue
            self._accept += extensions
            label = name_match.group('label')
            ext = name_match.group('ext')  # first extension, if multiple
            if len(label) == 0:
                label = f"{ext[1:].capitalize()} Files"
            extension_string = " ".join([f"*{ext}" for ext in extensions])
            filter_list.append(f"{label} ({extension_string})")
        self._filter = ";;".join(filter_list)

    def setBrowseButton(self, enable):
        self._btnBrowse.setVisible(enable)

    def setData(self, data):
        self._data = data

    def setDefaultPath(self, path):
        self._default_path = path

    def setFilter(self, f):
        self._filter = f

    def setText(self, filename):
        self._lineEdit.setText(filename)

    def text(self):
        return self._lineEdit.text()



import sys

from PyQt5 import QtCore

import xappt
import xappt_qt

from xappt_qt.gui.application import get_application
from xappt_qt.gui.ui.browser_tab_about import Ui_tabAbout
from xappt_qt.gui.tab_pages.base import BaseTabPage
from xappt_qt.constants import APP_TITLE

app_info = {
    "xappt_qt": {
        "version": xappt_qt.version_str,
        "url": "https://github.com/cmontesano/xappt_qt.git",
    },
    "xappt": {
        "version": xappt.version_str,
        "url": "https://github.com/cmontesano/xappt.git",
    },
    "python": {
        "version": ".".join(map(str, sys.version_info[:3])),
        "url": "https://www.python.org",
    },
    "PyQt": {
        "version": QtCore.PYQT_VERSION_STR,
        "url": "https://riverbankcomputing.com/software/pyqt/",
    },
}


class AboutTabPage(BaseTabPage, Ui_tabAbout):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.setupUi(self)
        self.lblTitle.setText(APP_TITLE)
        self.lblIcon.setPixmap(get_application().app_icon.pixmap(QtCore.QSize(48, 48)))
        self.txtAbout.setHtml(self.generate_html())

    @staticmethod
    def generate_html() -> str:
        html = ['<table cellspacing="4" cellpadding="4">']

        for name, app_data in app_info.items():
            url = app_data['url']
            version = app_data['version']
            html.append(f'<tr>'
                        f'<td><a href="{url}" style="text-decoration: none;">{name}</a></td>'
                        f'<td>{version}</td>'
                        f'</tr>')

        html.append("</table>")

        return "\n".join(html)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import pathlib

import setuptools

from xappt.utilities import git_tools, setup_helpers

PROJECT_PATH = pathlib.Path(__file__).absolute().parent
VERSION_PATH = PROJECT_PATH.joinpath("xappt_qt", "__version__.py")
README_PATH = PROJECT_PATH.joinpath("README.md")

os.chdir(PROJECT_PATH)


def main():
    if git_tools.is_dirty(PROJECT_PATH):
        print("Local repository is not clean")
        if not setup_helpers.ask("Proceed anyway?"):
            return

    if not setup_helpers.cleanup_build_files(PROJECT_PATH):
        return

    install_requires = list(setup_helpers.requirements(PROJECT_PATH))

    setup_dict = {
        'name': 'xappt_qt',
        'version': setup_helpers.get_version(VERSION_PATH),
        'author': 'Christopher Montesano',
        'author_email': 'cmdev.00+xappt@gmail.com',
        'url': 'https://github.com/cmontesano/xappt_qt',
        'description': 'Qt/PyQt5 interface plugin for xappt.',
        'long_description': README_PATH.read_text("utf8"),
        'long_description_content_type': 'text/markdown',
        'packages': setup_helpers.build_package_list('xappt_qt'),
        'include_package_data': True,
        'zip_safe': False,
        'license': 'MIT',
        'platforms': 'any',
        'python_requires': '>=3.7, <4',
        'install_requires': install_requires,
        'classifiers': [
            'Topic :: Utilities',
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
        ],
        'entry_points': {
            'console_scripts': [
                'xappt-qt=xappt_qt.main:main',
            ],
        },
    }

    commit_id = git_tools.commit_id(PROJECT_PATH, short=True)
    setup_helpers.update_build(VERSION_PATH, commit_id)
    setuptools.setup(**setup_dict)
    setup_helpers.update_build(VERSION_PATH, "dev")


if __name__ == '__main__':
    main()



from PyQt5 import QtWidgets

import xappt

from xappt_qt.gui.widgets.tool_page.converters.base import ParameterWidgetBase


class ParameterWidgetFloat(ParameterWidgetBase):
    def get_widget(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        w = QtWidgets.QDoubleSpinBox(parent=self)
        minimum = param.options.get("minimum", -999999999.0)
        maximum = param.options.get("maximum", 999999999.0)
        w.setMinimum(minimum)
        w.setMaximum(maximum)
        w.setStepType(QtWidgets.QDoubleSpinBox.AdaptiveDecimalStepType)
        for v in (param.value, param.default):
            if v is not None:
                w.setValue(v)
                break
        else:
            param.value = w.value()

        self._getter_fn = w.value
        self._setter_fn = w.setValue

        w.valueChanged[float].connect(lambda x: self.onValueChanged.emit(param.name, x))

        return w



from typing import Generator, Sequence

from PyQt5 import QtCore, QtWidgets

from xappt_qt.gui.delegates import SimpleItemDelegate


class CheckList(QtWidgets.QWidget):
    item_changed = QtCore.pyqtSignal(QtWidgets.QListWidgetItem)

    def __init__(self, searchable: bool = False):
        super().__init__()

        self.list = QtWidgets.QListWidget()
        self.setup_ui(searchable)

        self._init_context_menu()

    def setup_ui(self, searchable: bool):
        grid = QtWidgets.QGridLayout()
        grid.setContentsMargins(0, 0, 0, 0)
        self.setLayout(grid)

        list_row = 0
        if searchable:
            edit = QtWidgets.QLineEdit()
            edit.setPlaceholderText("Search")
            edit.setClearButtonEnabled(True)
            edit.textChanged.connect(self._on_filter_changed)
            grid.addWidget(edit, 0, 0)
            list_row = 1

        grid.addWidget(self.list, list_row, 0)

        self.list.setItemDelegate(SimpleItemDelegate())
        # noinspection PyUnresolvedReferences
        self.list.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.list.setAlternatingRowColors(True)
        self.list.setSpacing(2)
        self.list.itemChanged.connect(self._on_item_changed)
        self.setSizePolicy(self.list.sizePolicy())

    def _init_context_menu(self):
        self.list.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)

        self.context_menu = QtWidgets.QMenu()

        action_check = QtWidgets.QAction("Check All", self)
        action_check.setData(self.check_all)
        self.context_menu.addAction(action_check)

        action_uncheck = QtWidgets.QAction("Check None", self)
        action_uncheck.setData(self.uncheck_all)
        self.context_menu.addAction(action_uncheck)

        action_selected = QtWidgets.QAction("Check Selected", self)
        action_selected.setData(self.check_selected)
        self.context_menu.addAction(action_selected)

        action_invert = QtWidgets.QAction("Invert Checked", self)
        action_invert.setData(self.invert_checked)
        self.context_menu.addAction(action_invert)

        self.list.customContextMenuRequested.connect(self._on_context_menu)

    def add_item(self, text: str, state: QtCore.Qt.CheckState = QtCore.Qt.Unchecked):
        item = QtWidgets.QListWidgetItem(text)
        # noinspection PyTypeChecker
        item.setFlags(item.flags() | QtCore.Qt.ItemIsUserCheckable)
        item.setCheckState(state)
        self.list.addItem(item)

    def add_items(self, items: Sequence[str], state: QtCore.Qt.CheckState = QtCore.Qt.Unchecked):
        for item in items:
            self.add_item(item, state)

    def checked_items(self) -> Generator[str, None, None]:
        for i in range(self.list.count()):
            item = self.list.item(i)
            if item.checkState() == QtCore.Qt.Checked:
                yield item.text()

    def check_all(self):
        self._set_check_state(QtCore.Qt.Checked)

    def uncheck_all(self):
        self._set_check_state(QtCore.Qt.Unchecked)

    def invert_checked(self):
        for i in range(self.list.count()):
            item = self.list.item(i)
            if item.checkState() == QtCore.Qt.Checked:
                item.setCheckState(QtCore.Qt.Unchecked)
            else:
                item.setCheckState(QtCore.Qt.Checked)

    def check_selected(self):
        for i in range(self.list.count()):
            item = self.list.item(i)
            if item.isSelected():
                item.setCheckState(QtCore.Qt.Checked)
            else:
                item.setCheckState(QtCore.Qt.Unchecked)

    def check_items(self, items: Sequence[str]):
        for item in items:
            for match in self.list.findItems(item, QtCore.Qt.MatchExactly):
                match.setCheckState(QtCore.Qt.Checked)

    def _set_check_state(self, state: QtCore.Qt.CheckState):
        for i in range(self.list.count()):
            self.list.item(i).setCheckState(state)

    def _on_item_changed(self, item: QtWidgets.QListWidgetItem):
        self.item_changed.emit(item)

    def _on_context_menu(self, pos: QtCore.QPoint):
        if len(self.context_menu.actions()) == 0:
            return
        action = self.context_menu.exec_(self.list.mapToGlobal(pos))
        if action is None:
            return
        action_fn = action.data()
        action_fn()

    def _on_filter_changed(self, text: str):
        search_text = text.strip().lower()
        while search_text.count("  "):
            search_text = search_text.replace("  ", " ")
        search_terms = search_text.split()
        for i in range(self.list.count()):
            item = self.list.item(i)
            item_text = item.text().lower()
            matched_terms = True
            for term in search_terms:
                if term not in item_text:
                    matched_terms = False
                    break
            item.setHidden(not matched_terms)



import re

FILE_NAME_VALID_CHARS__RE = re.compile(r"\W")

__all__ = [
    'safe_file_name',
]


def safe_file_name(s: str, *, strip_underscores: bool = True) -> str:
    cleansed_string = FILE_NAME_VALID_CHARS__RE.sub("_", s.lower())
    while cleansed_string.count("__"):
        cleansed_string = cleansed_string.replace("__", "_")
    if strip_underscores:
        cleansed_string = cleansed_string.strip("_")
    return cleansed_string



from math import ceil

from PyQt5 import QtCore, QtGui, QtWidgets

from xappt_qt.gui.delegates.simple_item import SimpleItemDelegate


class ToolItemDelegate(SimpleItemDelegate):
    ROLE_TOOL_CLASS = QtCore.Qt.UserRole + 1
    ROLE_ITEM_TYPE = QtCore.Qt.UserRole + 2
    ROLE_ITEM_LIST = QtCore.Qt.UserRole + 3
    ROLE_ITEM_SEARCH_TEXT = QtCore.Qt.UserRole + 4
    ROLE_ICON_SIZE = QtCore.Qt.UserRole + 5

    ITEM_TYPE_COLLECTION = 0
    ITEM_TYPE_TOOL = 1

    def __init__(self, parent=None):
        super().__init__(parent)
        self.dark_bg_color = QtGui.QPalette().color(QtGui.QPalette.Highlight).darker(500)
        self.light_bg_color = QtGui.QPalette().color(QtGui.QPalette.Highlight)
        self.fg_color = QtGui.QPalette().color(QtGui.QPalette.ButtonText)
        self.outline_color = QtGui.QColor(0, 0, 0)
        self._hint_height = None

    def draw_arrow(self, expanded: bool, painter: QtGui.QPainter, option: QtWidgets.QStyleOptionViewItem):
        painter.save()

        painter.setRenderHints(QtGui.QPainter.Antialiasing, True)
        painter.setRenderHints(QtGui.QPainter.TextAntialiasing, True)
        painter.setRenderHints(QtGui.QPainter.HighQualityAntialiasing, True)

        y_scale = -1.0 if expanded else 1.0

        radius = option.rect.height() * 0.15
        arrow = QtGui.QPainterPath()
        arrow.moveTo(0.0 * radius, 1.0 * radius * y_scale)
        arrow.lineTo(0.866 * radius, -0.5 * radius * y_scale)
        arrow.lineTo(-0.866 * radius, -0.5 * radius * y_scale)
        arrow.lineTo(0.0 * radius, 1.0 * radius * y_scale)

        offset = option.rect.height() * 0.5
        arrow.translate(option.rect.right() - offset, option.rect.top() + offset)

        painter.setOpacity(0.5)
        for y in (-1, 0, 1):
            for x in (-1, 0, 1):
                painter.fillPath(arrow.translated(x, y), self.outline_color)

        painter.setOpacity(1.0)
        painter.fillPath(arrow, self.fg_color)
        painter.restore()

    def draw_bg(self, painter: QtGui.QPainter, option: QtWidgets.QStyleOptionViewItem):
        painter.save()

        if option.state & QtWidgets.QStyle.State_Selected == QtWidgets.QStyle.State_Selected:
            bg_color = self.light_bg_color
        else:
            bg_color = self.dark_bg_color

        painter.fillRect(option.rect, bg_color)
        painter.setPen(self.outline_color)
        painter.drawRect(option.rect.adjusted(0, 0, -1, -1))
        painter.restore()

    def draw_caption(self, caption: str, painter: QtGui.QPainter, option: QtWidgets.QStyleOptionViewItem):
        painter.save()

        painter.setRenderHints(QtGui.QPainter.Antialiasing, False)
        painter.setRenderHints(QtGui.QPainter.TextAntialiasing, False)
        painter.setRenderHints(QtGui.QPainter.HighQualityAntialiasing, False)

        font = painter.font()
        font.setBold(True)
        painter.setFont(font)

        painter.setPen(self.outline_color)
        painter.setOpacity(0.6)
        for y in (-1, 0, 1):
            for x in (-1, 0, 1):
                painter.drawText(option.rect.translated(x, y), QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter, caption)

        painter.setPen(self.fg_color)
        painter.setOpacity(1.0)
        painter.drawText(option.rect, QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter, caption)
        painter.restore()

    def paint(self, painter: QtGui.QPainter, option: QtWidgets.QStyleOptionViewItem, index: QtCore.QModelIndex):
        item_type = index.data(self.ROLE_ITEM_TYPE)

        if item_type == self.ITEM_TYPE_COLLECTION:
            painter.save()

            self.draw_bg(painter, option)

            if isinstance(option.widget, QtWidgets.QTreeView):
                self.draw_arrow(option.widget.isExpanded(index), painter, option)

            caption = f"  {index.data(QtCore.Qt.DisplayRole)}"
            self.draw_caption(caption, painter, option)
            painter.restore()
        else:
            super(ToolItemDelegate, self).paint(painter, option, index)

    def sizeHint(self, option: QtWidgets.QStyleOptionViewItem, index: QtCore.QModelIndex) -> QtCore.QSize:
        hint = super().sizeHint(option, index)
        # if self._hint_height is None:
        icon_size = index.data(self.ROLE_ICON_SIZE)
        if icon_size is not None:
            hint.setHeight(icon_size.height() + 12)
        return hint



from PyQt5 import QtCore

from xappt_qt.constants import ToolViewType
from xappt_qt.gui.ui.browser_tab_options import Ui_tabOptions
from xappt_qt.gui.tab_pages.base import BaseTabPage


class OptionsTabPage(BaseTabPage, Ui_tabOptions):
    options_changed = QtCore.pyqtSignal()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.setupUi(self)

        self.populate_tool_view_type_combo()

        self.chkMinimizeToTray.stateChanged.connect(self.options_changed.emit)
        self.chkStartMinimized.stateChanged.connect(self.options_changed.emit)
        self.cmbToolViewType.currentIndexChanged.connect(self.options_changed.emit)

    def populate_tool_view_type_combo(self):
        self.cmbToolViewType.clear()
        for value in ToolViewType:
            name: str = value.name
            if name.startswith("VIEW_"):
                name = name[5:]
            entry = name.replace("_", " ").title()
            self.cmbToolViewType.addItem(entry, value.value)

    def disable_tray_icon(self):
        self.chkMinimizeToTray.setChecked(False)
        self.chkMinimizeToTray.setEnabled(False)

    def settings(self) -> dict:
        return {
            'minimize_to_tray': self.chkMinimizeToTray.isChecked(),
            'start_minimized': self.chkStartMinimized.isChecked(),
            'view_type': self.cmbToolViewType.currentData(),
        }

    def apply_settings(self, settings_dict: dict):
        if self.chkMinimizeToTray.isEnabled():
            self.chkMinimizeToTray.setChecked(settings_dict.get('minimize_to_tray', True))
        else:
            self.chkMinimizeToTray.setChecked(False)
        self.chkStartMinimized.setChecked(settings_dict.get('start_minimized', False))
        view_type = settings_dict.get('view_type', ToolViewType.VIEW_SMALL_ICONS.value)
        index = max(0, self.cmbToolViewType.findData(view_type))
        self.cmbToolViewType.setCurrentIndex(index)



from PyQt5 import QtWidgets, QtCore

import xappt

from xappt_qt.gui.widgets.tool_page.converters.base import ParameterWidgetBase


class ParameterWidgetInt(ParameterWidgetBase):
    def get_widget(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        if param.choices is not None:
            return self._convert_int_choice(param)
        else:
            return self._convert_int(param)

    def _convert_int_choice(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        w = QtWidgets.QComboBox(parent=self)
        w.addItems(param.choices)
        for v in (param.value, param.default):
            if v is not None:
                if isinstance(v, str):
                    if v in param.choices:
                        index = w.findText(v)
                        w.setCurrentIndex(index)
                elif isinstance(v, int):
                    if 0 <= v < w.count():
                        w.setCurrentIndex(v)
                break
        else:
            param.value = w.currentIndex()

        self._setup_combobox(param, w)

        w.currentIndexChanged[str].connect(lambda x: self.onValueChanged.emit(param.name, x))

        self._getter_fn = w.currentIndex
        self._setter_fn = w.setCurrentIndex

        return w

    def _convert_int(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        minimum = param.options.get("minimum")
        maximum = param.options.get("maximum")

        ui = param.options.get("ui")
        if ui == "slider" and minimum is not None and maximum is not None:
            w = QtWidgets.QSlider(parent=self)
            w.setOrientation(QtCore.Qt.Horizontal)
            ticks = param.options.get("ticks", 0)
            if ticks == 0:
                w.setTickPosition(QtWidgets.QSlider.NoTicks)
            else:
                w.setTickPosition(QtWidgets.QSlider.TicksBelow)
                w.setTickInterval(ticks)
        else:
            w = QtWidgets.QSpinBox(parent=self)
            if minimum is None:
                minimum = -999999999
            if maximum is None:
                maximum = 999999999
        w.setMinimum(minimum)
        w.setMaximum(maximum)
        for v in (param.value, param.default):
            if v is not None:
                w.setValue(v)
                break
        else:
            param.value = w.value()
        w.valueChanged[int].connect(lambda x: self.onValueChanged.emit(param.name, x))

        self._getter_fn = w.value
        self._setter_fn = w.setValue

        return w



# -*- coding: utf-8 -*-


#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_tabOptions(object):
    def setupUi(self, tabOptions):
        tabOptions.setObjectName("tabOptions")
        tabOptions.resize(263, 300)
        self.verticalLayout = QtWidgets.QVBoxLayout(tabOptions)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gridLayout_3 = QtWidgets.QGridLayout()
        self.gridLayout_3.setContentsMargins(6, 6, 6, 6)
        self.gridLayout_3.setSpacing(12)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.cmbToolViewType = QtWidgets.QComboBox(tabOptions)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cmbToolViewType.sizePolicy().hasHeightForWidth())
        self.cmbToolViewType.setSizePolicy(sizePolicy)
        self.cmbToolViewType.setObjectName("cmbToolViewType")
        self.gridLayout_3.addWidget(self.cmbToolViewType, 2, 1, 1, 1)
        self.chkMinimizeToTray = QtWidgets.QCheckBox(tabOptions)
        self.chkMinimizeToTray.setObjectName("chkMinimizeToTray")
        self.gridLayout_3.addWidget(self.chkMinimizeToTray, 0, 1, 1, 1)
        self.chkStartMinimized = QtWidgets.QCheckBox(tabOptions)
        self.chkStartMinimized.setObjectName("chkStartMinimized")
        self.gridLayout_3.addWidget(self.chkStartMinimized, 1, 1, 1, 1)
        self.label = QtWidgets.QLabel(tabOptions)
        self.label.setObjectName("label")
        self.gridLayout_3.addWidget(self.label, 2, 0, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout_3)
        spacerItem = QtWidgets.QSpacerItem(20, 237, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)

        self.retranslateUi(tabOptions)
        QtCore.QMetaObject.connectSlotsByName(tabOptions)
        tabOptions.setTabOrder(self.chkMinimizeToTray, self.chkStartMinimized)
        tabOptions.setTabOrder(self.chkStartMinimized, self.cmbToolViewType)

    def retranslateUi(self, tabOptions):
        _translate = QtCore.QCoreApplication.translate
        tabOptions.setWindowTitle(_translate("tabOptions", "Options"))
        self.chkMinimizeToTray.setText(_translate("tabOptions", "Minimize to system tray"))
        self.chkStartMinimized.setText(_translate("tabOptions", "Start Minimized"))
        self.label.setText(_translate("tabOptions", "View"))



import xappt


@xappt.register_plugin
class CustomIconAutoModule(xappt.BaseTool):
    custom_icon = "icon2.svg"

    @classmethod
    def name(cls) -> str:
        return "custom-icon-auto-module"

    @classmethod
    def help(cls) -> str:
        return ("This tool will display a custom icon in the **xappt_qt browser** and the **Tool dialog**.\n "
                "A class variable named `custom_icon` is set with a string value representing an image file name.\n "
                "The module containing this tool's class will automatically derived, and will be used in conjunction "
                "with the image file name to retrieve a file path using `importlib.resources`.\n\n " 
                "*Note that the tool dialog will only display the custom icon of the first loaded tool.*")

    @classmethod
    def collection(cls) -> str:
        return "Examples"

    def execute(self, **kwargs) -> int:
        self.interface.message("Complete")
        return 0
